#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/tooltest2.py: Test the 'mmgen-tool' utility
"""

# TODO: move all non-interactive 'mmgen-tool' tests in 'cmdtest.py' here
# TODO: move all(?) tests in 'tooltest.py' here (or duplicate them?)

import sys, os, time, importlib, asyncio
from subprocess import run, PIPE

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from test.include.common import set_globals, end_msg, init_coverage

from mmgen import main_tool
from mmgen.cfg import Config
from mmgen.color import green, blue, purple, cyan, gray
from mmgen.util import msg, msg_r, Msg, die

skipped_tests = ['mn2hex_interactive']
coin_dependent_groups = ('Coin', 'File')

opts_data = {
	'text': {
		'desc': "Simple test suite for the 'mmgen-tool' utility",
		'usage':'[options] [command]...',
		'options': """
-h, --help           Print this help message
-a, --no-altcoin     Skip altcoin tests
-A, --tool-api       Test the tool_api subsystem
-C, --coverage       Produce code coverage info using trace module
-d, --die-on-missing Abort if no test data found for given command
--, --longhelp       Print help message for long (global) options
-l, --list-tests     List the test groups in this test suite
-L, --list-tested-cmds Output the 'mmgen-tool' commands that are tested by this test suite
-n, --names          Print command names instead of descriptions
-q, --quiet          Produce quieter output
-t, --type=          Specify coin type
-f, --fork           Run commands via tool executable instead of importing tool module
-v, --verbose        Produce more verbose output
""",
	'notes': """

If no command is given, the whole suite of tests is run.
"""
	}
}

sys.argv = [sys.argv[0]] + ['--skip-cfg-file'] + sys.argv[1:]

cfg = Config(
	opts_data = opts_data,
	init_opts = {
		'usr_randchars': 0,
		'hash_preset': '1',
		'passwd_file': 'test/ref/keyaddrfile_password',
	})

set_globals(cfg)

from test.tooltest2_d.data import *

def fork_cmd(cmd_name, args, opts, stdin_input):
	cmd = (
		tool_cmd_preargs +
		tool_cmd +
		(opts or []) +
		[cmd_name] + args
	)
	vmsg('{} {}'.format(
		green('Executing'),
		cyan(' '.join(cmd))))
	cp = run(cmd, input=stdin_input or None, stdout=PIPE, stderr=PIPE)
	try:
		cmd_out = cp.stdout.decode()
	except:
		cmd_out = cp.stdout
	if cp.stderr:
		vmsg(cp.stderr.strip().decode())
	if cp.returncode != 0:
		import re
		m = re.search(b'tool command returned (None|False)', cp.stderr)
		if m:
			return eval(m.group(1))
		else:
			die(2, f'Spawned program exited with error: {cp.stderr}')

	return cmd_out.strip()

def call_method(cls, method, cmd_name, args, mmtype, stdin_input):
	vmsg('{a}: {b}{c}'.format(
		a = purple('Running'),
		b = ' '.join([cmd_name]+[repr(e) for e in args]),
		c = ' '+mmtype if mmtype else ''))
	aargs, kwargs = main_tool.process_args(cmd_name, args, cls)
	oq_save = bool(cfg.quiet)
	if not cfg.verbose:
		cfg._set_quiet(True)
	if stdin_input:
		fd0, fd1 = os.pipe()
		if os.fork(): # parent
			os.close(fd1)
			stdin_save = os.dup(0)
			os.dup2(fd0, 0)
			cmd_out = method(*aargs, **kwargs)
			os.dup2(stdin_save, 0)
			os.wait()
			cfg._set_quiet(oq_save)
			return cmd_out
		else: # child
			os.close(fd0)
			os.write(fd1, stdin_input)
			vmsg(f'Input: {stdin_input!r}')
			sys.exit(0)
	else:
		ret = method(*aargs, **kwargs)
		if type(ret).__name__ == 'coroutine':
			ret = asyncio.run(ret)
		cfg._set_quiet(oq_save)
		return ret

def tool_api(cls, cmd_name, args, opts):
	from mmgen.tool.api import tool_api
	tool = tool_api(cfg)
	if opts:
		for o in opts:
			if o.startswith('--type='):
				tool.addrtype = o.split('=')[1]
	pargs, kwargs = main_tool.process_args(cmd_name, args, cls)
	return getattr(tool, cmd_name)(*pargs, **kwargs)

def check_output(out, chk):
	if isinstance(chk, str):
		chk = chk.encode()
	if isinstance(out, int):
		out = str(out).encode()
	if isinstance(out, str):
		out = out.encode()
	err_fs = "Output ({!r}) doesn't match expected output ({!r})"
	try:
		outd = out.decode()
	except:
		outd = None

	if type(chk).__name__ == 'function':
		assert chk(outd), f'{chk.__name__}({outd}) failed!'
	elif isinstance(chk, dict):
		for k, v in chk.items():
			if k == 'boolfunc':
				assert v(outd), f'{v.__name__}({outd}) failed!'
			elif k == 'value':
				assert outd == v, err_fs.format(outd, v)
			else:
				outval = getattr(__builtins__, k)(out)
				if outval != v:
					die(1, f'{k}({out}) returned {outval}, not {v}!')
	elif chk is not None:
		assert out == chk, err_fs.format(out, chk)

def run_test(cls, gid, cmd_name):
	data = tests[gid][cmd_name]

	# behavior is like cmdtest.py: run coin-dependent tests only if proto.testnet or proto.coin != BTC
	if gid in coin_dependent_groups:
		k = '{}_{}'.format(
			(cfg.token.lower() if proto.tokensym else proto.coin.lower()),
			proto.network)
		if k in data:
			data = data[k]
			m2 = f' ({k})'
		else:
			qmsg(f'-- no data for {cmd_name} ({k}) - skipping')
			return
	else:
		if proto.coin != 'BTC' or proto.testnet:
			return
		m2 = ''

	m = '{} {}{}'.format(
		purple('Testing'),
		cmd_name if cfg.names else docstring_head(getattr(cls, cmd_name)),
		m2)

	msg_r(green(m)+'\n' if cfg.verbose else m)
	skipping = False

	for n, d in enumerate(data):
		args, out, opts, mmtype = d + tuple([None] * (4-len(d)))
		if 'fmt=xmrseed' in args and cfg.no_altcoin:
			if not skipping:
				qmsg('')
			skip_msg = f'Skipping altcoin test {cmd_name} {args}'
			qmsg(('' if n else '\n') + gray(skip_msg if len(skip_msg) <= 100 else skip_msg[:97] + '...'))
			skipping = True
			continue
		skipping = False
		stdin_input = None
		if args and isinstance(args[0], bytes):
			stdin_input = args[0]
			args[0] = '-'

		if cfg.tool_api:
			if args and args[0]== '-':
				continue
			cmd_out = tool_api(cls, cmd_name, args, opts)
		elif cfg.fork:
			cmd_out = fork_cmd(cmd_name, args, opts, stdin_input)
		else:
			if stdin_input and sys.platform == 'win32':
				msg(gray('Skipping for MSWin - no os.fork()'))
				continue
			method = getattr(cls(cfg, cmdname=cmd_name, proto=proto, mmtype=mmtype), cmd_name)
			cmd_out = call_method(cls, method, cmd_name, args, mmtype, stdin_input)

		try:
			vmsg(f'Output:\n{cmd_out}\n')
		except:
			vmsg(f'Output:\n{cmd_out!r}\n')

		if isinstance(out, tuple) and type(out[0]).__name__ == 'function':
			func_out = out[0](cmd_out)
			assert func_out == out[1], (
				'{}({}) == {} failed!\nOutput: {}'.format(
					out[0].__name__,
					cmd_out,
					out[1],
					func_out))
		elif isinstance(out, (list, tuple)):
			for co, o in zip(cmd_out.split(NL) if cfg.fork else cmd_out, out):
				check_output(co, o)
		else:
			check_output(cmd_out, out)

		if not cfg.verbose:
			msg_r('.')

	if not cfg.verbose:
		msg('OK')

def docstring_head(obj):
	return obj.__doc__.strip().split('\n')[0] if obj.__doc__ else None

def do_group(gid):
	desc = f'command group {gid!r}'
	cls = main_tool.get_mod_cls(gid.lower())
	qmsg(blue('Testing ' +
		desc if cfg.names else
		(docstring_head(cls) or desc)
	))

	for cmdname in cls(cfg).user_commands:
		if cmdname in skipped_tests:
			continue
		if cmdname not in tests[gid]:
			m = f'No test for command {cmdname!r} in group {gid!r}!'
			if cfg.die_on_missing:
				die(1, m+'  Aborting')
			else:
				msg(m)
				continue
		run_test(cls, gid, cmdname)

def do_cmd_in_group(cmdname):
	cls = main_tool.get_cmd_cls(cmdname)
	for gid, cmds in tests.items():
		for cmd in cmds:
			if cmd == cmdname:
				run_test(cls, gid, cmdname)
				return True
	return False

def list_tested_cmds():
	for gid in tests:
		Msg('\n'.join(tests[gid]))

def main():
	if cfg._args:
		for cmd in cfg._args:
			if cmd in tests:
				do_group(cmd)
			else:
				if not do_cmd_in_group(cmd):
					die(1, f'{cmd!r}: not a recognized test or test group')
	else:
		for garg in tests:
			do_group(garg)

qmsg = cfg._util.qmsg
vmsg = cfg._util.vmsg

proto = cfg._proto

if cfg.tool_api:
	del tests['Wallet']
	del tests['File']

if cfg.list_tests:
	Msg('Available tests:')
	for modname, cmdlist in main_tool.mods.items():
		cls = getattr(importlib.import_module(f'mmgen.tool.{modname}'), 'tool_cmd')
		Msg(f'  {modname:6} - {docstring_head(cls)}')
	sys.exit(0)

if cfg.list_tested_cmds:
	list_tested_cmds()
	sys.exit(0)

tool_exec = os.path.relpath(os.path.join('cmds', 'mmgen-tool'))

if cfg.fork:
	passthru_args = ['coin', 'type', 'testnet', 'token']
	tool_cmd = [tool_exec, '--skip-cfg-file'] + [
		'--{}{}'.format(
			k.replace('_', '-'),
			'='+getattr(cfg, k) if getattr(cfg, k) is not True else '')
		for k in passthru_args if getattr(cfg, k)]

	if cfg.coverage:
		d, f = init_coverage()
		tool_cmd_preargs = ['python3', '-m', 'trace', '--count', '--coverdir='+d, '--file='+f]
	else:
		tool_cmd_preargs = ['python3', 'scripts/exec_wrapper.py']

from mmgen.main import launch
start_time = int(time.time())
launch(func=main)
end_msg(int(time.time()) - start_time)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
main: Script launcher for the MMGen Project
"""

import sys, os

def launch(*, mod=None, func=None, fqmod=None, package='mmgen'):

	if sys.platform in ('linux', 'darwin') and sys.stdin.isatty():
		import termios, atexit
		fd = sys.stdin.fileno()
		old = termios.tcgetattr(fd)
		atexit.register(lambda: termios.tcsetattr(fd, termios.TCSADRAIN, old))

	try:
		__import__(f'{package}.main_{mod}') if mod else func() if func else __import__(fqmod)
	except KeyboardInterrupt:
		from .color import yellow
		sys.stderr.write(yellow('\nUser interrupt\n'))
		sys.exit(1)
	except EOFError:
		from .color import yellow
		sys.stderr.write(yellow('\nEnd of file\n'))
		sys.exit(1)
	except Exception as e:
		try:
			errmsg = str(e.args[0])
		except:
			errmsg = repr(e.args[0]) if e.args else repr(e)

		from collections import namedtuple
		from .color import nocolor, yellow, red

		_o = namedtuple('exit_data', ['color', 'exit_val', 'fs'])
		d = {
			0:   _o(nocolor, 0, '{message}'),
			1:   _o(nocolor, 1, '{message}'),
			2:   _o(yellow,  2, '{message}'),
			3:   _o(yellow,  3, '\nMMGen Error ({name}):\n{message}'),
			4:   _o(red,     4, '\nMMGen Fatal Error ({name}):\n{message}'),
			'x': _o(yellow,  5, '\nMMGen Python Exception ({name}): {e}'),
		}[getattr(e, 'mmcode', 'x')]

		(sys.stdout if getattr(e, 'stdout', None) else sys.stderr).write(
			d.color(d.fs.format(
				name = type(e).__name__,
				message = errmsg.strip() or e,
				e = e))
			+ '\n')

		if os.getenv('MMGEN_EXEC_WRAPPER') or os.getenv('MMGEN_TRACEBACK'):
			raise

		sys.exit(d.exit_val)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.rpc: JSON/RPC routines for the 'mmgen-tool' utility
"""

from .common import tool_cmd_base, options_annot_str
from ..tw.view import TwView
from ..tw.txhistory import TwTxHistory

class tool_cmd(tool_cmd_base):
	"tracking-wallet commands using the JSON-RPC interface"

	need_proto = True
	need_amt = True

	async def daemon_version(self):
		"print coin daemon version"
		from ..daemon import CoinDaemon
		d = CoinDaemon(cfg=self.cfg, proto=self.proto, test_suite=self.cfg.test_suite)
		if self.proto.base_proto == 'Monero':
			from ..proto.xmr.rpc import MoneroRPCClient
			r = MoneroRPCClient(
				cfg    = self.cfg,
				proto  = self.proto,
				daemon = d,
				host   = 'localhost',
				port   = d.rpc_port,
				user   = None,
				passwd = None,
				ignore_daemon_version = True)
		else:
			from ..rpc import rpc_init
			r = await rpc_init(self.cfg, self.proto, ignore_daemon_version=True, ignore_wallet=True)
		return f'{d.coind_name} version {r.daemon_version} ({r.daemon_version_str})'

	async def getbalance(self, *,
			minconf: 'minimum number of confirmations' = 1,
			quiet:   'produce quieter output' = False,
			pager:   'send output to pager' = False):
		"list confirmed/unconfirmed, spendable/unspendable balances in tracking wallet"
		from ..tw.bal import TwGetBalance
		return (await TwGetBalance(
			self.cfg, self.proto, minconf=minconf, quiet=quiet)).format(color=self.cfg.color)

	async def twops(self,
			obj, pager, reverse, detail, sort, age_fmt, interactive,
			**kwargs):

		obj.reverse = reverse
		obj.age_fmt = age_fmt

		for k, v in kwargs.items():
			setattr(obj, k, v)

		await obj.get_data(sort_key=sort, reverse_sort=reverse)

		if interactive:
			await obj.view_filter_and_sort()
			ret = True
		else:
			ret = await obj.format('detail' if detail else 'squeezed')

		if hasattr(obj, 'twctl'):
			del obj.twctl

		return ret

	async def twview(self, *,
			pager:       'send output to pager' = False,
			reverse:     'reverse order of unspent outputs' = False,
			wide:        'display data in wide tabular format' = False,
			minconf:     'minimum number of confirmations' = 1,
			sort:        'unspent output sort order ' + options_annot_str(TwView.sort_funcs) = 'age',
			age_fmt:     'format for the Age/Date column ' + options_annot_str(TwView.age_fmts) = 'confs',
			interactive: 'enable interactive operation' = False,
			show_mmid:   'show MMGen IDs along with coin addresses' = True):
		"view tracking wallet unspent outputs"

		from ..tw.unspent import TwUnspentOutputs
		obj = await TwUnspentOutputs(self.cfg, self.proto, minconf=minconf)
		return await self.twops(
			obj, pager, reverse, wide, sort, age_fmt, interactive,
			show_mmid = show_mmid)

	async def txhist(self, *,
			pager:       'send output to pager' = False,
			reverse:     'reverse order of transactions' = False,
			detail:      'produce detailed, non-tabular output' = False,
			sinceblock:  'display transactions starting from this block' = 0,
			sort:        'transaction sort order ' + options_annot_str(TwTxHistory.sort_funcs) = 'age',
			age_fmt:     'format for the Age/Date column ' + options_annot_str(TwView.age_fmts) = 'confs',
			interactive: 'enable interactive operation' = False):
		"view transaction history of tracking wallet"

		obj = await TwTxHistory(self.cfg, self.proto, sinceblock=sinceblock)
		return await self.twops(
			obj, pager, reverse, detail, sort, age_fmt, interactive)

	async def listaddress(self,
			mmgen_addr: str,
			*,
			wide:         'display data in wide tabular format' = False,
			minconf:      'minimum number of confirmations' = 1,
			showcoinaddr: 'display coin address in addition to MMGen ID' = True,
			age_fmt:      'format for the Age/Date column ' + options_annot_str(TwView.age_fmts) = 'confs'):
		"list the specified MMGen address in the tracking wallet and its balance"

		return await self.listaddresses(
			mmgen_addrs   = mmgen_addr,
			wide          = wide,
			minconf       = minconf,
			showcoinaddrs = showcoinaddr,
			age_fmt       = age_fmt)

	async def listaddresses(self, *,
			pager:        'send output to pager' = False,
			reverse:      'reverse order of unspent outputs' = False,
			wide:         'display data in wide tabular format' = False,
			minconf:      'minimum number of confirmations' = 1,
			sort:         'address sort order ' + options_annot_str(['reverse', 'mmid', 'addr', 'amt']) = '',
			age_fmt:      'format for the Age/Date column ' + options_annot_str(TwView.age_fmts) = 'confs',
			interactive:  'enable interactive operation' = False,
			mmgen_addrs:  'hyphenated range or comma-separated list of addresses' = '',
			showcoinaddrs:'display coin addresses in addition to MMGen IDs' = True,
			showempty:    'show addresses with no balances' = True,
			showused:     'show used addresses (tristate: 0=no, 1=yes, 2=all)' = 1,
			all_labels:   'show all addresses with labels' = False):
		"list MMGen addresses in the tracking wallet and their balances"

		assert showused in (0, 1, 2), "‘showused’ must have a value of 0, 1 or 2"

		from ..tw.addresses import TwAddresses
		obj = await TwAddresses(self.cfg, self.proto, minconf=minconf, mmgen_addrs=mmgen_addrs)
		return await self.twops(
			obj, pager, reverse, wide, sort, age_fmt, interactive,
			showcoinaddrs = showcoinaddrs,
			showempty     = showempty,
			showused      = showused,
			all_labels    = all_labels)

	async def add_label(self, mmgen_or_coin_addr: str, label: str):
		"add descriptive label for address in tracking wallet"
		from ..tw.ctl import TwCtl
		return await (await TwCtl(self.cfg, self.proto, mode='w')).set_comment(mmgen_or_coin_addr, label)

	async def remove_label(self, mmgen_or_coin_addr: str):
		"remove descriptive label for address in tracking wallet"
		await self.add_label(mmgen_or_coin_addr, '')
		return True

	async def remove_address(self, mmgen_or_coin_addr: str):
		"remove an address from tracking wallet"
		from ..tw.ctl import TwCtl
		# returns None on failure:
		ret = await (await TwCtl(self.cfg, self.proto, mode='w')).remove_address(mmgen_or_coin_addr)
		if ret:
			from ..util import msg
			msg(f'Address {ret!r} deleted from tracking wallet')
		return ret

	async def resolve_address(self, mmgen_or_coin_addr: str):
		"resolve an MMGen address in the tracking wallet to a coin address or vice-versa"
		from ..tw.ctl import TwCtl
		ret = await (await TwCtl(self.cfg, self.proto, mode='w')).resolve_address(mmgen_or_coin_addr)
		if ret:
			from ..addr import is_coin_addr
			return ret.twmmid if is_coin_addr(self.proto, mmgen_or_coin_addr) else ret.coinaddr
		else:
			return False

	async def rescan_address(self, mmgen_or_coin_addr: str):
		"rescan an address in the tracking wallet to update its balance"
		from ..tw.ctl import TwCtl
		return await (await TwCtl(self.cfg, self.proto, mode='w')).rescan_address(mmgen_or_coin_addr)

	async def rescan_blockchain(self, *,
			start_block: int = None,
			stop_block: int  = None):
		"""
		rescan the blockchain to update historical transactions in the tracking wallet

		NOTE:

		  The rescanning process typically takes several hours and may be interrupted
		  using Ctrl-C.  An interrupted rescan may be resumed using the ‘start_block’
		  parameter.
		"""
		from ..tw.ctl import TwCtl
		await (await TwCtl(self.cfg, self.proto, mode='w')).rescan_blockchain(start_block, stop_block)
		return True

	async def twexport(self, *,
			include_amts = True,
			pretty       = False,
			prune        = False,
			warn_used    = False,
			force        = False):
		"""
		export a tracking wallet to JSON format

		NOTES:

		  If ‘include_amts’ is true (the default), Ethereum balances will be restored
		  from the dump upon import. For Bitcoin and forks, amount fields in the dump
		  are ignored.

		  If ‘pretty’ is true, JSON will be dumped in human-readable format to allow
		  for editing of comment fields.

		  If ‘prune’ is true, an interactive menu will be launched allowing the user
		  to prune unwanted addresses before creating the JSON dump.  Pruning has no
		  effect on the existing tracking wallet.

		  If ‘warn_used’ is true, the user will be prompted before pruning used
		  addresses.

		  If ‘force’ is true, any existing dump will be overwritten without prompting.
		"""
		from ..tw.json import TwJSON
		await TwJSON.Export(
			self.cfg,
			self.proto,
			include_amts    = include_amts,
			pretty          = pretty,
			prune           = prune,
			warn_used       = warn_used,
			force_overwrite = force)
		return True

	async def twimport(self, filename: str, *, ignore_checksum=False, batch=False):
		"""
		restore a tracking wallet from a JSON dump created by ‘twexport’

		NOTES:

		  If comment fields in the JSON dump have been edited, ‘ignore_checksum’ must
		  be set to true.

		  The restored tracking wallet will have correct balances but no record of
		  historical transactions.  These may be restored by running ‘mmgen-tool
		  rescan_blockchain’.
		"""
		from ..tw.json import TwJSON
		await TwJSON.Import(self.cfg, self.proto, filename, ignore_checksum=ignore_checksum, batch=batch)
		return True



# Copyright (c) 2021 Pieter Wuille
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
#
# Source: https://github.com/bitcoin/bitcoin/blob/master/test/functional/test_framework/ripemd160.py
#
# Adapted for the MMGen Project with the following changes:
# - replace leading spaces with tabs
# - reimplement as class with digest() and hexdigest() methods
# - add custom test output display
#
# SECURITY DISCLAIMER:
#   This code has been flagged as ‘test-only’ by its creator because it is not
#   constant-time and therefore vulnerable to timing attacks.
#
#   The RIPEMD-160 algorithm is used in the MMGen code base in two places: a)
#   in hash160(); and b) as a checksum for a config file template.  Here
#   timing attacks are not meaningful, because in the first case the preimage
#   is a secure cryptographic hash and in the second -- public data.
#
#   The MMGen Project assumes no responsibility for the use of this code in
#   other contexts.

"""Test-only pure Python RIPEMD160 implementation."""

# Message schedule indexes for the left path.
ML = [
	0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
	7, 4, 13, 1, 10, 6, 15, 3, 12, 0, 9, 5, 2, 14, 11, 8,
	3, 10, 14, 4, 9, 15, 8, 1, 2, 7, 0, 6, 13, 11, 5, 12,
	1, 9, 11, 10, 0, 8, 12, 4, 13, 3, 7, 15, 14, 5, 6, 2,
	4, 0, 5, 9, 7, 12, 2, 10, 14, 1, 3, 8, 11, 6, 15, 13
]

# Message schedule indexes for the right path.
MR = [
	5, 14, 7, 0, 9, 2, 11, 4, 13, 6, 15, 8, 1, 10, 3, 12,
	6, 11, 3, 7, 0, 13, 5, 10, 14, 15, 8, 12, 4, 9, 1, 2,
	15, 5, 1, 3, 7, 14, 6, 9, 11, 8, 12, 2, 10, 0, 4, 13,
	8, 6, 4, 1, 3, 11, 15, 0, 5, 12, 2, 13, 9, 7, 10, 14,
	12, 15, 10, 4, 1, 5, 8, 7, 6, 2, 13, 14, 0, 3, 9, 11
]

# Rotation counts for the left path.
RL = [
	11, 14, 15, 12, 5, 8, 7, 9, 11, 13, 14, 15, 6, 7, 9, 8,
	7, 6, 8, 13, 11, 9, 7, 15, 7, 12, 15, 9, 11, 7, 13, 12,
	11, 13, 6, 7, 14, 9, 13, 15, 14, 8, 13, 6, 5, 12, 7, 5,
	11, 12, 14, 15, 14, 15, 9, 8, 9, 14, 5, 6, 8, 6, 5, 12,
	9, 15, 5, 11, 6, 8, 13, 12, 5, 12, 13, 14, 11, 8, 5, 6
]

# Rotation counts for the right path.
RR = [
	8, 9, 9, 11, 13, 15, 15, 5, 7, 7, 8, 11, 14, 14, 12, 6,
	9, 13, 15, 7, 12, 8, 9, 11, 7, 7, 12, 7, 6, 15, 13, 11,
	9, 7, 15, 11, 8, 6, 6, 14, 12, 13, 5, 14, 13, 13, 7, 5,
	15, 5, 8, 11, 14, 14, 6, 14, 6, 9, 12, 9, 12, 5, 15, 8,
	8, 5, 12, 9, 12, 5, 14, 6, 8, 13, 6, 5, 15, 13, 11, 11
]

# K constants for the left path.
KL = [0, 0x5a827999, 0x6ed9eba1, 0x8f1bbcdc, 0xa953fd4e]

# K constants for the right path.
KR = [0x50a28be6, 0x5c4dd124, 0x6d703ef3, 0x7a6d76e9, 0]


def fi(x, y, z, i):
	"""The f1, f2, f3, f4, and f5 functions from the specification."""
	if i == 0:
		return x ^ y ^ z
	elif i == 1:
		return (x & y) | (~x & z)
	elif i == 2:
		return (x | ~y) ^ z
	elif i == 3:
		return (x & z) | (y & ~z)
	elif i == 4:
		return x ^ (y | ~z)
	else:
		assert False


def rol(x, i):
	"""Rotate the bottom 32 bits of x left by i bits."""
	return ((x << i) | ((x & 0xffffffff) >> (32 - i))) & 0xffffffff


def compress(h0, h1, h2, h3, h4, block):
	"""Compress state (h0, h1, h2, h3, h4) with block."""
	# Left path variables.
	al, bl, cl, dl, el = h0, h1, h2, h3, h4
	# Right path variables.
	ar, br, cr, dr, er = h0, h1, h2, h3, h4
	# Message variables.
	x = [int.from_bytes(block[4*i:4*(i+1)], 'little') for i in range(16)]

	# Iterate over the 80 rounds of the compression.
	for j in range(80):
		rnd = j >> 4
		# Perform left side of the transformation.
		al = rol(al + fi(bl, cl, dl, rnd) + x[ML[j]] + KL[rnd], RL[j]) + el
		al, bl, cl, dl, el = el, al, bl, rol(cl, 10), dl
		# Perform right side of the transformation.
		ar = rol(ar + fi(br, cr, dr, 4 - rnd) + x[MR[j]] + KR[rnd], RR[j]) + er
		ar, br, cr, dr, er = er, ar, br, rol(cr, 10), dr

	# Compose old state, left transform, and right transform into new state.
	return h1 + cl + dr, h2 + dl + er, h3 + el + ar, h4 + al + br, h0 + bl + cr

class ripemd160:

	def __init__(self,data=b''):
		"""Compute the RIPEMD-160 hash of data."""
		# Initialize state.
		state = (0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476, 0xc3d2e1f0)
		# Process full 64-byte blocks in the input.
		for b in range(len(data) >> 6):
			state = compress(*state, data[64*b:64*(b+1)])
		# Construct final blocks (with padding and size).
		pad = b"\x80" + b"\x00" * ((119 - len(data)) & 63)
		fin = data[len(data) & ~63:] + pad + (8 * len(data)).to_bytes(8, 'little')
		# Process final blocks.
		for b in range(len(fin) >> 6):
			state = compress(*state, fin[64*b:64*(b+1)])
		# Produce output.
		self.res = b"".join((h & 0xffffffff).to_bytes(4, 'little') for h in state)

	def hexdigest(self):
		return self.res.hex()

	def digest(self):
		return self.res

	def update(self,*args,**kwargs):
		raise NotImplementedError('update() method not implemented for pure-Python ripemd160')

	def copy(self,*args,**kwargs):
		raise NotImplementedError('copy() method not implemented for pure-Python ripemd160')

if __name__ == '__main__':

	# See https://homes.esat.kuleuven.be/~bosselae/ripemd160.html
	vectors = [
		(b"", "9c1185a5c5e9fc54612808977ee8f548b2258d31"),
		(b"a", "0bdc9d2d256b3ee9daae347be6f4dc835a467ffe"),
		(b"abc", "8eb208f7e05d987a9b044a8e98c6b087f15a0bfc"),
		(b"message digest", "5d0689ef49d2fae572b881b123a85ffa21595f36"),
		(b"abcdefghijklmnopqrstuvwxyz",
			"f71c27109c692c1b56bbdceb5b9d2865b3708dbc"),
		(b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
			"12a053384a9c0c88e405a06c27dcf49ada62eb2b"),
		(b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789",
			"b0e20b6e3116640286ed3a87a5713079b21f5189"),
		(b"1234567890" * 8, "9b752e45573d4b39f4dbd3323cab82bf63326bfb"),
		(b"a" * 1000000, "52783243c1697bdbe16d37f97f68f08325dc1528")
	]

	def test_ripemd160():
		import sys
		verbose = [s for s in ('verbose','--verbose') if s in sys.argv]
		fast = [s for s in ('fast','--fast') if s in sys.argv]
		sys.stderr.write("Testing RIPEMD-160 test vectors...")
		sys.stderr.flush()

		fs = '{:<7} {:62}{:3} {}\n'
		col1_w = 62
		if verbose:
			sys.stderr.write('\n' + fs.format('BYTES','INPUT','','OUTPUT'))

		for msg,hexout in (vectors[:-1] if fast else vectors):

			a = ripemd160(msg).hexdigest()
			b = hexout
			assert a == b, f'{a} != {b}'

			if verbose:
				sys.stderr.write(fs.format(
					len(msg),
					msg.decode()[:col1_w],
					"..." if len(msg) > col1_w else "",
					hexout ))

		sys.stderr.write('OK\n')

	test_ripemd160()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.cfg: swap configuration class for the MMGen Wallet suite
"""

import re
from collections import namedtuple

from ..amt import UniAmt
from ..util import die

_mmd = namedtuple('swap_config_option', ['min', 'max', 'dfl'])

class SwapCfg:

	# The trade limit, i.e., set 100000000 to get a minimum of 1 full asset, else a refund
	# Optional. 1e8 or scientific notation
	trade_limit = None

	# Swap interval for streaming swap in blocks. Optional. If 0, do not stream
	si = _mmd(1, 20, 3) # stream_interval

	# Swap quantity for streaming swap.
	# The interval value determines the frequency of swaps in blocks
	# Optional. If 0, network will determine the number of swaps
	stream_quantity = 0

	def __init__(self, cfg):

		self.cfg = cfg

		if cfg.trade_limit is not None:
			self.set_trade_limit(desc='parameter for --trade-limit')

		if cfg.stream_interval is None:
			self.stream_interval = self.si.dfl
		else:
			self.set_stream_interval(desc='parameter for --stream-interval')

	def set_trade_limit(self, *, desc):
		s = self.cfg.trade_limit
		if re.match(r'-*[0-9]+(\.[0-9]+)*%*$', s):
			self.trade_limit = 1 - float(s[:-1]) / 100 if s.endswith('%') else UniAmt(s)
		else:
			die('SwapCfgValueError', f'{s}: invalid {desc}')

	def set_stream_interval(self, *, desc):
		s = self.cfg.stream_interval
		from ..util import is_int
		if not is_int(s):
			die('SwapCfgValueError', f'{s}: invalid {desc} (not an integer)')
		self.stream_interval = si = int(s)
		if si < self.si.min:
			die('SwapCfgValueError', f'{si}: invalid {desc} (< {self.si.min})')
		if si > self.si.max:
			die('SwapCfgValueError', f'{si}: invalid {desc} (> {self.si.max})')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.include.pexpect: pexpect implementation for MMGen Wallet cmdtest suite
"""

import sys, time

from mmgen.color import red, yellow, green, cyan
from mmgen.util import msg, msg_r, rmsg, die

from ...include.common import cfg, vmsg, vmsg_r, getrandstr, strip_ansi_escapes

try:
	import pexpect
	from pexpect.popen_spawn import PopenSpawn
except ImportError as e:
	die(2, red(f'‘pexpect’ module is missing.  Cannnot run test suite ({e!r})'))

def debug_pexpect_msg(p):
	msg('\n{}{}{}'.format(red('BEFORE ['), p.before, red(']')))
	msg('{}{}{}'.format(red('MATCH ['), p.after, red(']')))

NL = '\n'

class CmdTestPexpect:

	def __init__(
			self,
			args,
			no_output     = False,
			spawn_env     = None,
			pexpect_spawn = False,
			send_delay    = None,
			timeout       = None,
			silent        = False,
			direct_exec   = False):

		self.pexpect_spawn = pexpect_spawn
		self.send_delay = send_delay
		self.skip_ok = False
		self.sent_value = None
		self.spawn_env = spawn_env
		self.exit_val = None

		if direct_exec or cfg.direct_exec:
			from subprocess import Popen, DEVNULL
			redir = DEVNULL if (no_output or not cfg.exact_output) else None
			self.ep = Popen([args[0]] + args[1:], stderr=redir, env=spawn_env)
		else:
			timeout = int(
				timeout
				or cfg.pexpect_timeout
				or cfg.test_suite_pexpect_timeout) or (60, 5)[bool(cfg.debug_pexpect)]
			if pexpect_spawn:
				self.p = pexpect.spawn(args[0], args[1:], encoding='utf8', timeout=timeout, env=spawn_env)
			else:
				self.p = PopenSpawn(args, encoding='utf8', timeout=timeout, env=spawn_env)

			if cfg.exact_output and not silent:
				self.p.logfile = sys.stdout

	def do_decrypt_ka_data(
			self,
			pw,
			desc         = 'key-address data',
			check        = True,
			have_yes_opt = False):
		self.passphrase(desc, pw)
		if not have_yes_opt:
			self.expect('Check key-to-address validity? (y/N): ', ('n', 'y')[check])

	def view_tx(self, view):
		self.expect(r'View.* transaction.*\? .*: ', view, regex=True)
		if cfg.pexpect_spawn and view == 'v':
			self.expect('END', 'q')
		elif view not in 'vn\n':
			self.expect('to continue: ', '\n')

	def do_comment(self, add_comment, has_label=False):
		p = ('Add a comment to transaction', 'Edit transaction comment')[has_label]
		self.expect(f'{p}? (y/N): ', ('n', 'y')[bool(add_comment)])
		if add_comment:
			self.expect('Comment: ', add_comment+'\n')

	def ok(self, exit_val=None):
		if not self.pexpect_spawn:
			self.p.sendeof()
		self.p.read()
		ret = self.p.wait()
		if ret != (self.exit_val or exit_val or 0) and not cfg.coverage:
			die('TestSuiteSpawnedScriptException', f'Spawned script exited with value {ret}')
		if cfg.profile:
			return
		if not self.skip_ok:
			m = 'OK\n' if ret == 0 else f'OK[{ret}]\n'
			sys.stderr.write(green(m) if cfg.exact_output or cfg.verbose else ' '+m)
		return self

	def license(self):
		if self.spawn_env.get('MMGEN_NO_LICENSE'):
			return
		self.expect("'w' for conditions and warranty info, or 'c' to continue: ", 'c')

	def label(self, label='Test Label (UTF-8) α'):
		self.expect('Enter a wallet label, or hit ENTER for no label: ', label+'\n')

	def usr_rand(self, num_chars):
		if cfg.usr_random:
			self.interactive()
			self.send('\n')
		else:
			rand_chars = list(getrandstr(num_chars, no_space=True))
			vmsg_r('SEND ')
			while rand_chars:
				ch = rand_chars.pop(0)
				msg_r(yellow(ch)+' ' if cfg.verbose else '+')
				self.expect('left: ', ch, delay=0.005)
			self.expect('ENTER to continue: ', '\n')

	def passphrase_new(self, desc, passphrase):
		self.expect(f'Enter passphrase for {desc}: ', passphrase+'\n')
		self.expect('Repeat passphrase: ', passphrase+'\n')

	def passphrase(self, desc, passphrase, pwtype=''):
		if pwtype:
			pwtype += ' '
		self.expect(f'Enter {pwtype}passphrase for {desc}.*?: ', passphrase+'\n', regex=True)

	def hash_preset(self, desc, preset=''):
		self.expect(f'Enter hash preset for {desc}')
		self.expect('or hit ENTER .*?:', str(preset)+'\n', regex=True)

	def written_to_file(self, desc, overwrite_unlikely=False, query='Overwrite?  '):
		s1 = f'{desc} written to file '
		s2 = query + "Type uppercase 'YES' to confirm: "
		ret = self.expect(([s1, s2], s1)[overwrite_unlikely])
		if ret == 1:
			self.send('YES\n')
			return self.expect_getend("Overwriting file '").rstrip("'")
		self.expect(NL, nonl=True)
		outfile = self.p.before.strip().strip("'")
		if cfg.debug_pexpect:
			rmsg(f'Outfile [{outfile}]')
		vmsg('{} file: {}'.format(desc, cyan(outfile.replace('"', ""))))
		return outfile

	def hincog_create(self, hincog_bytes):
		ret = self.expect(['Create? (Y/n): ', "'YES' to confirm: "])
		if ret == 0:
			self.send('\n')
			self.expect('Enter file size: ', str(hincog_bytes)+'\n')
		else:
			self.send('YES\n')
		return ret

	def no_overwrite(self):
		self.expect("Overwrite?  Type uppercase 'YES' to confirm: ", '\n')
		self.expect('Exiting at user request')

	def expect_getend(self, s, regex=False):
		self.expect(s, regex=regex, nonl=True)
		if cfg.debug_pexpect:
			debug_pexpect_msg(self.p)
		# readline() of partial lines doesn't work with PopenSpawn, so do this instead:
		self.expect(NL, nonl=True, silent=True)
		if cfg.debug_pexpect:
			debug_pexpect_msg(self.p)
		end = self.p.before.rstrip()
		if not cfg.debug:
			vmsg(f' ==> {cyan(end)}')
		return end

	def interactive(self):
		return self.p.interact() # interact() not available with popen_spawn

	def kill(self, signal):
		return self.p.kill(signal)

	def match_expect_list(self, expect_list, greedy=False):
		allrep = '.*' if greedy else '.*?'
		expect = (
			r'(\b|\s)' +
			fr'\s{allrep}\s'.join(s.replace(r'.', r'\.').replace(' ', r'\s+') for s in expect_list) +
			r'(\b|\s)')
		import re
		m = re.search(expect, self.read(strip_color=True), re.DOTALL)
		assert m, f'No match found for regular expression {expect!r}'
		return m

	def expect(self, s, t='', delay=None, regex=False, nonl=False, silent=False):

		if not silent:
			if cfg.verbose:
				msg_r('EXPECT ' + yellow(str(s)))
			elif not cfg.exact_output:
				msg_r('+')

		try:
			ret = (self.p.expect_exact, self.p.expect)[bool(regex)](s) if s else 0
		except pexpect.TIMEOUT as e:
			if cfg.debug_pexpect:
				raise
			m1 = f'\nERROR.  Expect {s!r} timed out.  Exiting\n'
			m2 = f'before: [{self.p.before}]\n'
			m3 = f'sent value: [{self.sent_value}]' if self.sent_value is not None else ''
			raise pexpect.TIMEOUT(m1+m2+m3) from e

		if cfg.debug_pexpect:
			debug_pexpect_msg(self.p)

		if cfg.verbose and not isinstance(s, str):
			msg_r(f' ==> {ret} ')

		if ret == -1:
			die(4, f'Error.  Expect returned {ret}')
		else:
			if t:
				self.send(t, delay, s)
			else:
				if not nonl and not silent:
					vmsg('')
			return ret

	def send(self, t, delay=None, s=False):
		delay = delay or self.send_delay
		if delay:
			time.sleep(delay)
			if cfg.demo:
				time.sleep(0.5)
		ret = self.p.send(t) # returns num bytes written
		self.sent_value = t if ret else None
		if cfg.demo and delay:
			time.sleep(delay)
		if cfg.verbose:
			ls = '' if cfg.debug or not s else ' '
			es = '' if s else '  '
			yt = yellow('{!r}'.format(t.replace('\n', r'\n')))
			msg(f'{ls}SEND {es}{yt}')
		return ret

	def read(self, n=-1, strip_color=False):
		return strip_ansi_escapes(self.p.read(n)).replace('\r', '') if strip_color else self.p.read(n)

	def close(self):
		if self.pexpect_spawn:
			self.p.close()



#!/usr/bin/env python3

import os

from mmgen.cfg import Config
from mmgen.util import msg

opts_data = {
	'sets': [('print_checksum', True, 'quiet', True)],
	'text': {
		'desc': 'Opts test',
		'usage':'[args] [opts]',
		'options': """
-h, --help            Print this help message
--, --longhelp        Print help message for long (global) options
-i, --in-fmt=      f  Input is from wallet format 'f'
-d, --outdir=      d  Use outdir 'd'
-C, --print-checksum  Print a checksum
-E, --fee-estimate-mode=M Specify the network fee estimate mode.
-F, --no-foobleize    Do not foobleize the output, even on user request
-H, --hidden-incog-input-params=f,o  Read hidden incognito data from file
                      'f' at offset 'o' (comma-separated)
-k, --keep-label      Reuse label of input wallet for output wallet
-l, --seed-len=    l  Specify wallet seed length of 'l' bits.
-L, --label=       l  Specify a label 'l' for output wallet
-m, --minconf=     n  Minimum number of confirmations required to spend
                      outputs (default: 1)
-o, --show-opts=   L  List of cfg opts to display
-p, --hash-preset= p  Use the scrypt hash parameters defined by preset 'p'
-P, --passwd-file= f  Get wallet passphrase from file 'f'
-q, --quiet           Be quieter
-t, --min-temp=    t  Minimum temperature (in degrees Celsius)
-T, --max-temp=    t  Maximum temperature (in degrees Celsius)
-x, --point=       P  Point in Euclidean space
-X, --cached-balances Use cached balances (Ethereum only)
-v, --verbose         Be more verbose
                      sample help_note: {kgs}
                      sample help_note: {coin_id}
""",
	'notes': """

                           NOTES FOR THIS COMMAND

sample note: {nn}
"""
	},
	'code': {
		'options': lambda cfg, help_notes, s: s.format(
			kgs     = help_notes('keygen_backends'),
			coin_id = help_notes('coin_id'),
		),
		'notes': lambda s: s.format(nn='a note'),
	}
}

cfg = Config(opts_data=opts_data, need_proto=os.getenv('TEST_MISC_OPTS_NEEDS_PROTO'))

if cfg.show_opts:
	opts = cfg.show_opts.split(',')
	col1_w = max(len(s) for s in opts) + 5
	for opt in opts:
		msg('{:{w}} {}'.format(f'cfg.{opt}:', getattr(cfg, opt), w=col1_w))
		if cfg._proto:
			coin, *rem = opt.split('_')
			network = rem[0] if rem[0] in cfg._proto.network_names else None
			opt_name = '_'.join(rem[bool(network):])
			msg('{:{w}} {}'.format(f'proto.{opt_name}:', getattr(cfg._proto, opt_name), w=col1_w))

msg('')
for n, arg in enumerate(cfg._args, 1):
	msg(f'arg{n}: {arg}')



import os as overlay_fake_os
from .rpc_orig import *

if overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA'):

	class overlay_fake_data:
		rpc_init = rpc_init

	async def rpc_init(*args, **kwargs):
		from .obj import NonNegativeInt
		ret = await overlay_fake_data.rpc_init(*args, **kwargs)
		ret.blockcount = NonNegativeInt(1000000)
		return ret



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.txcreate_examples: txcreate and txdo help examples for the MMGen Wallet suite
"""

from ..cfg import gc

def help(proto, cfg):

	mmtype = 'B' if 'B' in proto.mmtypes else proto.mmtypes[0]
	from ..tool.coin import tool_cmd
	t = tool_cmd(cfg, mmtype=mmtype)
	addr = t.privhex2addr('bead' * 16)
	sample_addr = addr.views[addr.view_pref]

	return f"""
EXAMPLES:

  Send 0.123 {proto.coin} to an external {proto.name} address, returning the change to a
  specific MMGen address in the tracking wallet:

    $ {gc.prog_name} {sample_addr},0.123 01ABCDEF:{mmtype}:7

  Same as above, but select the change address automatically:

    $ {gc.prog_name} {sample_addr},0.123 01ABCDEF:{mmtype}

  Same as above, but select the change address automatically by address type:

    $ {gc.prog_name} {sample_addr},0.123 {mmtype}

  Same as above, but reduce verbosity and specify fee of 20 satoshis
  per byte:

    $ {gc.prog_name} -q -f 20s {sample_addr},0.123 {mmtype}

  Send entire balance of selected inputs minus fee to an external {proto.name}
  address:

    $ {gc.prog_name} {sample_addr}

  Send entire balance of selected inputs minus fee to first unused wallet
  address of specified type:

    $ {gc.prog_name} {mmtype}
""" if proto.base_proto == 'Bitcoin' else f"""
EXAMPLES:

  Send 0.123 {proto.coin} to an external {proto.name} address:

    $ {gc.prog_name} {sample_addr},0.123

  Send 0.123 {proto.coin} to another account in wallet 01ABCDEF:

    $ {gc.prog_name} 01ABCDEF:{mmtype}:7,0.123
"""



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.incog_hidden: hidden incognito wallet class
"""

import sys, os

from ..seed import Seed
from ..util import msg, die, capfirst
from ..util2 import parse_bytespec
from .incog_base import wallet

class wallet(wallet):

	desc = 'hidden incognito data'
	file_mode = 'binary'
	no_tty = True

	_msg = {
		'choose_file_size': """
  You must choose a size for your new hidden incog data.  The minimum size
  is {} bytes, which puts the incog data right at the end of the file.
  Since you probably want to hide your data somewhere in the middle of the
  file where it's harder to find, you're advised to choose a much larger file
  size than this.
	""",
		'check_incog_id': """
  Check generated Incog ID above against your records.  If it doesn't match,
  then your incognito data is incorrect or corrupted, or you may have speci-
  fied an incorrect offset.
	""",
		'record_incog_id': """
  Make a record of the Incog ID but keep it secret.  You will used it to
  identify the incog wallet data in the future and to locate the offset
  where the data is hidden in the event you forget it.
	""",
		'decrypt_params': ', hash preset, offset {} seed length'
	}

	def _get_hincog_params(self, wtype):
		a = getattr(self.cfg, 'hidden_incog_'+ wtype +'_params').split(',')
		return ','.join(a[:-1]), int(a[-1]) # permit comma in filename

	def _check_valid_offset(self, fn, action):
		d = self.ssdata
		m = ('Input', 'Destination')[action=='write']
		if fn.size < d.hincog_offset + d.target_data_len:
			die(1, '{a} file {b!r} has length {c}, too short to {d} {e} bytes of data at offset {f}'.format(
				a = m,
				b = fn.name,
				c = fn.size,
				d = action,
				e = d.target_data_len,
				f = d.hincog_offset))

	def _get_data(self):
		d = self.ssdata
		d.hincog_offset = self._get_hincog_params('input')[1]

		self.cfg._util.qmsg(f'Getting hidden incog data from file {self.infile.name!r}')

		# Already sanity-checked:
		d.target_data_len = self._get_incog_data_len(self.cfg.seed_len or Seed.dfl_len)
		self._check_valid_offset(self.infile, 'read')

		flgs = os.O_RDONLY|os.O_BINARY if sys.platform == 'win32' else os.O_RDONLY
		fh = os.open(self.infile.name, flgs)
		os.lseek(fh, int(d.hincog_offset), os.SEEK_SET)
		self.fmt_data = os.read(fh, d.target_data_len)
		os.close(fh)
		self.cfg._util.qmsg(f'Data read from file {self.infile.name!r} at offset {d.hincog_offset}')

	# overrides method in Wallet
	def write_to_file(self, **kwargs):
		d = self.ssdata
		self._format()
		self.cfg._util.compare_or_die(
			val1  = d.target_data_len,
			desc1 = 'target data length',
			val2  = len(self.fmt_data),
			desc2 = 'length of formatted ' + self.desc)

		k = ('output', 'input')[self.op=='pwchg_new']
		fn, d.hincog_offset = self._get_hincog_params(k)

		if self.cfg.outdir and not os.path.dirname(fn):
			fn = os.path.join(self.cfg.outdir, fn)

		check_offset = True
		try:
			os.stat(fn)
		except:
			from ..ui import keypress_confirm, line_input
			keypress_confirm(
				self.cfg,
				f'Requested file {fn!r} does not exist.  Create?',
				default_yes = True,
				do_exit = True)

			min_fsize = d.target_data_len + d.hincog_offset
			msg('\n  ' + self.msg['choose_file_size'].strip().format(min_fsize)+'\n')
			while True:
				fsize = parse_bytespec(line_input(self.cfg, 'Enter file size: '))
				if fsize >= min_fsize:
					break
				msg(f'File size must be an integer no less than {min_fsize}')

			from ..tool.fileutil import tool_cmd
			tool_cmd(self.cfg).rand2file(fn, str(fsize))
			check_offset = False

		from ..filename import MMGenFile
		f = MMGenFile(fn, subclass=type(self), write=True)

		self.cfg._util.dmsg('{} data len {}, offset {}'.format(
			capfirst(self.desc),
			d.target_data_len,
			d.hincog_offset))

		if check_offset:
			self._check_valid_offset(f, 'write')
			if not self.cfg.quiet:
				from ..ui import confirm_or_raise
				confirm_or_raise(
					self.cfg,
					message = '',
					action  = f'alter file {f.name!r}')

		flgs = os.O_RDWR|os.O_BINARY if sys.platform == 'win32' else os.O_RDWR
		fh = os.open(f.name, flgs)
		os.lseek(fh, int(d.hincog_offset), os.SEEK_SET)
		os.write(fh, self.fmt_data)
		os.close(fh)
		msg('{} written to file {!r} at offset {}'.format(
			capfirst(self.desc),
			f.name,
			d.hincog_offset))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
addrdata: MMGen AddrData and related classes
"""

from .cfg import gc
from .util import fmt, die
from .base_obj import AsyncInit
from .obj import MMGenObject, MMGenDict, get_obj
from .addr import MMGenID, AddrListID
from .addrlist import AddrListEntry, AddrListData, AddrList

class AddrData(MMGenObject):

	def __init__(self, proto):
		self.al_ids = {}
		self.proto = proto
		self.rpc = None

	def seed_ids(self):
		return list(self.al_ids.keys())

	def addrlist(self, al_id):
		# TODO: Validate al_id
		if al_id in self.al_ids:
			return self.al_ids[al_id]

	def mmaddr2coinaddr(self, mmaddr):
		al_id, idx = MMGenID(self.proto, mmaddr).rsplit(':', 1)
		coinaddr = ''
		if al_id in self.al_ids:
			coinaddr = self.addrlist(al_id).coinaddr(int(idx))
		return coinaddr or None

	def coinaddr2mmaddr(self, coinaddr):
		d = self.make_reverse_dict([coinaddr])
		return (list(d.values())[0][0]) if d else None

	def add(self, addrlist):
		if isinstance(addrlist, AddrList):
			self.al_ids[addrlist.al_id] = addrlist
			return True
		else:
			raise TypeError(f'Error: object {addrlist!r} is not an instance of AddrList')

	def make_reverse_dict(self, coinaddrs):
		d = MMGenDict()
		for al_id in self.al_ids:
			d.update(self.al_ids[al_id].make_reverse_dict_addrlist(coinaddrs))
		return d

class TwAddrData(AddrData, metaclass=AsyncInit):

	def __new__(cls, cfg, proto, *, twctl=None):
		return MMGenObject.__new__(proto.base_proto_subclass(cls, 'addrdata'))

	async def __init__(self, cfg, proto, *, twctl=None):
		from .rpc import rpc_init
		from .tw.shared import TwLabel
		from .seed import SeedID
		self.cfg = cfg
		self.proto = proto
		self.rpc = await rpc_init(cfg, proto)
		self.al_ids = {}
		twd = await self.get_tw_data(twctl=twctl)
		out, i = {}, 0
		for acct, addr_array in twd:
			l = get_obj(TwLabel, proto=self.proto, text=acct, silent=True)
			if l and l.mmid.type == 'mmgen':
				obj = l.mmid.obj
				if len(addr_array) != 1:
					message = self.msgs['multiple_acct_addrs'].strip().format(acct=acct, proj=gc.proj_name)
					die(3, fmt(message, indent='  '))
				al_id = AddrListID(
					sid = SeedID(sid=obj.sid),
					mmtype = self.proto.addr_type(obj.mmtype))
				if al_id not in out:
					out[al_id] = []
				out[al_id].append(AddrListEntry(self.proto, idx=obj.idx, addr=addr_array[0], comment=l.comment))
				i += 1

		self.cfg._util.vmsg(f'{i} {gc.proj_name} addresses found, {len(twd)} accounts total')

		for al_id in out:
			self.add(AddrList(
				self.cfg,
				self.proto,
				al_id = al_id,
				adata = AddrListData(sorted(out[al_id], key=lambda a: a.idx))
			))

class TwAddrDataWithStore(TwAddrData):

	msgs = {
		'multiple_acct_addrs': """
			ERROR: More than one address found for account: {acct!r}.
			Your tracking wallet is corrupted!
		"""
	}

	async def get_tw_data(self, *, twctl=None):
		self.cfg._util.vmsg('Getting address data from tracking wallet')
		if twctl is None:
			from .tw.ctl import TwCtl
			twctl = await TwCtl(self.cfg, self.proto)
		# emulate the output of RPC 'listaccounts' and 'getaddressesbyaccount'
		return [(mmid+' '+d['comment'], [d['addr']]) for mmid, d in list(twctl.mmid_ordered_dict.items())]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.include.ecc: elliptic curve utilities for the MMGen test suite
"""

import ecdsa, hashlib
from mmgen.proto.secp256k1.keygen import pubkey_format

def _pubkey_to_pub_point(vk_bytes):
	try:
		return ecdsa.VerifyingKey.from_string(vk_bytes, curve=ecdsa.curves.SECP256k1).pubkey.point
	except Exception as e:
		raise ValueError(f'invalid pubkey {vk_bytes.hex()}\n    {type(e).__name__}: {e}')

def _check_pub_point(pub_point, vk_bytes, addend_bytes=None):
	if pub_point is ecdsa.ellipticcurve.INFINITY:
		raise ValueError(
			'pubkey {}{} produced key with point at infinity!'.format(
				vk_bytes.hex(),
				'' if addend_bytes is None else f' + {addend_bytes.hex()}'))

def pubkey_check_pyecdsa(vk_bytes):
	_check_pub_point(_pubkey_to_pub_point(vk_bytes), vk_bytes)

def pubkey_tweak_add_pyecdsa(vk_bytes, pk_addend_bytes):
	pk_addend = int.from_bytes(pk_addend_bytes, byteorder='big')
	point_sum = (
		_pubkey_to_pub_point(vk_bytes) +
		ecdsa.SigningKey.from_secret_exponent(pk_addend, curve=ecdsa.SECP256k1).verifying_key.pubkey.point
	)
	_check_pub_point(point_sum, vk_bytes, pk_addend_bytes)
	return pubkey_format(
		ecdsa.VerifyingKey.from_public_point(point_sum, curve=ecdsa.curves.SECP256k1).to_string(),
		compressed = len(vk_bytes) == 33)

def sign_msghash_pyecdsa(msghash, privkey):
	ec_privkey = ecdsa.SigningKey.from_string(privkey, curve=ecdsa.curves.SECP256k1)
	return ec_privkey.sign_digest_deterministic(
		msghash,
		hashfunc = hashlib.sha256,
		sigencode = ecdsa.util.sigencode_string_canonize)

def verify_sig_pyecdsa(sig, msghash, pubkey):
	ec_pubkey = ecdsa.VerifyingKey.from_string(pubkey, curve=ecdsa.curves.SECP256k1)
	return ec_pubkey.verify_digest(sig, msghash)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
cfgfile: API for the MMGen runtime configuration file and related files
"""

import os, re
from collections import namedtuple

from .cfg import gc
from .util import msg, ymsg, suf, fmt, fmt_list, oneshot_warning, strip_comment, capfirst, die

def mmgen_cfg_file(cfg, id_str):
	return cfg_file.get_cls_by_id(id_str)(cfg)

class cfg_file:
	cur_ver = 2
	ver = None
	write_ok = False
	warn_missing = True
	write_metadata = False
	line_data = namedtuple('cfgfile_line', ['name', 'value', 'lineno', 'chunk'])
	fn_base = 'mmgen.cfg'

	class warn_missing_file(oneshot_warning):
		color = 'yellow' # has no effect, as color not initialized yet
		message = '{} not found at {!r}'

	def get_data(self, fn):
		try:
			with open(fn) as fp:
				return fp.read().splitlines()
		except:
			if self.warn_missing:
				self.warn_missing_file(div=fn, fmt_args=(self.desc, fn))
			return ''

	def copy_system_data(self, fn):
		assert self.write_ok, f'writing to file {fn!r} not allowed!'
		src = mmgen_cfg_file(self.cfg, 'sys')
		if src.data:
			data = src.data + src.make_metadata() if self.write_metadata else src.data
			try:
				with open(fn, 'w') as fp:
					fp.write('\n'.join(data)+'\n')
				os.chmod(fn, 0o600)
			except:
				die(2, f'ERROR: unable to write to {fn!r}')

	def parse_value(self, value, refval):
		if isinstance(refval, dict):
			m = re.fullmatch(r'((\s+\w+:\S+)+)', ' '+value) # expect one or more colon-separated values
			if m:
				return dict([i.split(':') for i in m[1].split()])
		elif isinstance(refval, (list, tuple)):
			m = re.fullmatch(r'((\s+\S+)+)', ' '+value)     # expect single value or list
			if m:
				ret = m[1].split()
				return ret if isinstance(refval, list) else tuple(ret)
		else:
			return value

	def get_lines(self):
		def gen_lines():
			for lineno, line in enumerate(self.data, 1):
				line = strip_comment(line)
				if line == '':
					continue
				m = re.fullmatch(r'(\w+)(\s+)(.*)', line)
				if m:
					yield self.line_data(m[1], m[3], lineno, None)
				else:
					die('CfgFileParseError', f'Parse error in file {self.fn!r}, line {lineno}')
		return gen_lines()

	@classmethod
	def get_cls_by_id(cls, id_str):
		d = {
			'usr':    CfgFileUsr,
			'sys':    CfgFileSampleSys,
			'sample': CfgFileSampleUsr,
		}
		return d[id_str]

class cfg_file_sample(cfg_file):

	@classmethod
	def cls_make_metadata(cls, data):
		return [f'# Version {cls.cur_ver} {cls.compute_chksum(data)}']

	@staticmethod
	def compute_chksum(data):
		import hashlib
		return hashlib.new('ripemd160', '\n'.join(data).encode()).hexdigest()

	@property
	def computed_chksum(self):
		return type(self).compute_chksum(self.data)

	def get_lines(self):
		"""
		The config file template contains some 'magic':
		- lines must either be empty or begin with '# '
		- each commented chunk must end with a parsable cfg variable line
		- chunks are delimited by one or more blank lines
		- lines beginning with '##' are ignored
		- everything up to first line beginning with '##' is ignored
		- last line is metadata line of the form '# Version VER_NUM HASH'
		"""

		def process_chunk(chunk, lineno):
			m = re.fullmatch(r'(#\s*)(\w+)(\s+)(.*)', chunk[-1])
			if m:
				return self.line_data(m[2], m[4], lineno, chunk)
			else:
				die('CfgFileParseError', f'Parse error in file {self.fn!r}, line {lineno}')

		def gen_chunks(lines):
			hdr = True
			chunk = []
			in_chunk = False
			last_nonblank = 0

			for lineno, line in enumerate(lines, 1):

				if line.startswith('##'):
					hdr = False
					continue

				if hdr:
					continue

				if line == '':
					in_chunk = False
				elif line.startswith('#'):
					if in_chunk is False:
						if chunk:
							yield process_chunk(chunk, last_nonblank)
						chunk = [line]
						in_chunk = True
					else:
						chunk.append(line)
					last_nonblank = lineno
				else:
					die('CfgFileParseError', f'Parse error in file {self.fn!r}, line {lineno}')

			if chunk:
				yield process_chunk(chunk, last_nonblank)

		return list(gen_chunks(self.data))

class CfgFileUsr(cfg_file):
	desc = 'user configuration file'
	warn_missing = False
	write_ok = True

	def __init__(self, cfg):
		self.cfg = cfg
		self.fn = os.path.join(cfg.data_dir_root, self.fn_base)
		self.data = self.get_data(self.fn)
		if not self.data:
			self.copy_system_data(self.fn)

class CfgFileSampleSys(cfg_file_sample):
	desc = 'system sample configuration file'
	test_fn_subdir = 'usr.local.share'

	def __init__(self, cfg):
		self.cfg = cfg
		if self.cfg.test_suite_cfgtest:
			self.fn = os.path.join(cfg.data_dir_root, self.test_fn_subdir, self.fn_base)
			with open(self.fn) as fp:
				self.data = fp.read().splitlines()
		else:
			# self.fn is used for error msgs only, so file need not exist on filesystem
			self.fn = os.path.join(os.path.dirname(__file__), 'data', self.fn_base)
			self.data = gc.get_mmgen_data_file(filename=self.fn_base).splitlines()

	def make_metadata(self):
		return [f'# Version {self.cur_ver} {self.computed_chksum}']

class CfgFileSampleUsr(cfg_file_sample):
	desc = 'sample configuration file'
	warn_missing = False
	write_ok = True
	chksum = None
	write_metadata = True
	details_confirm_prompt = 'View details?'
	out_of_date_fs = 'File {!r} is out of date - replacing'
	altered_by_user_fs = 'File {!r} was altered by user - replacing'

	def __init__(self, cfg):
		self.cfg = cfg
		self.fn = os.path.join(cfg.data_dir_root, f'{self.fn_base}.sample')
		self.data = self.get_data(self.fn)

		src = mmgen_cfg_file(cfg, 'sys')

		if not src.data:
			return

		if self.data:
			if self.parse_metadata():
				if self.chksum == self.computed_chksum:
					diff = self.diff(self.get_lines(), src.get_lines())
					if not diff:
						return
					self.show_changes(diff)
				else:
					msg(self.altered_by_user_fs.format(self.fn))
			else:
				msg(self.out_of_date_fs.format(self.fn))

		self.copy_system_data(self.fn)

	def parse_metadata(self):
		if self.data:
			m = re.match(r'# Version (\d+) ([a-f0-9]{40})$', self.data[-1])
			if m:
				self.ver = m[1]
				self.chksum = m[2]
				self.data = self.data[:-1] # remove metadata line
				return True

	def diff(self, a_tup, b_tup): # a=user, b=system
		a = [i.name for i in a_tup]#[3:] # Debug
		b = [i.name for i in b_tup]#[:-2] # Debug
		removed = set(a) - set(b)
		added   = set(b) - set(a)
		if removed or added:
			return {
				'removed': [i for i in a_tup if i.name in removed],
				'added':   [i for i in b_tup if i.name in added],
			}
		else:
			return None

	def show_changes(self, diff):
		ymsg('Warning: configuration file options have changed!\n')
		for desc in ('added', 'removed'):
			changed_opts = [i.name for i in diff[desc]
				# workaround for coin-specific opts previously listed in sample file:
				if not (i.name.endswith('_ignore_daemon_version') and desc == 'removed')
			]
			if changed_opts:
				msg(f'  The following option{suf(changed_opts, verb="has")} been {desc}:')
				msg(f'    {fmt_list(changed_opts, fmt="bare")}\n')
				if desc == 'removed':
					uc = mmgen_cfg_file(self.cfg, 'usr')
					usr_names = [i.name for i in uc.get_lines()]
					bad = sorted(set(usr_names).intersection(changed_opts))
					if bad:
						m = f"""
							The following removed option{suf(bad, verb='is')} set in {uc.fn!r}
							and must be deleted or commented out:
							{'  ' + fmt_list(bad, fmt='bare')}
						"""
						ymsg(fmt(m, indent='  ', strip_char='\t'))

		from .ui import keypress_confirm, do_pager
		while True:
			if not keypress_confirm(self.cfg, self.details_confirm_prompt, no_nl=True):
				return

			def get_details():
				for desc, data in diff.items():
					sep, sep2 = ('\n  ', '\n\n  ')
					if data:
						yield (
							f'{capfirst(desc)} section{suf(data)}:'
							+ sep2
							+ sep2.join([f'{sep.join(v.chunk)}' for v in data])
						)

			do_pager(
				'CONFIGURATION FILE CHANGES\n\n'
				+ '\n\n'.join(get_details()) + '\n'
			)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.cosmos.tx.protobuf: transaction serialization for Cosmos protocol
"""

# cosmjs-types/src/cosmos/bank/v1beta1/tx.ts
# cosmjs-types/src/cosmos/tx/v1beta1/tx.ts
# cosmjs-types/src/cosmos/tx/signing/v1beta1/signing.ts

from hashlib import sha256

from dataclasses import dataclass
from pure_protobuf.annotations import Field
from pure_protobuf.message import BaseMessage
from typing import Annotated, List, Optional
from enum import IntEnum

class SignMode(IntEnum):
	SIGN_MODE_UNSPECIFIED = 0
	SIGN_MODE_DIRECT = 1
	SIGN_MODE_TEXTUAL = 2
	SIGN_MODE_DIRECT_AUX = 3
	SIGN_MODE_LEGACY_AMINO_JSON = 127
	SIGN_MODE_EIP_191 = 191
	UNRECOGNIZED = -1

@dataclass
class PublicKeyData(BaseMessage):
	data: Annotated[bytes, Field(1)]

@dataclass
class PublicKey(BaseMessage):
	id:  Annotated[str, Field(1)] #  '/cosmos.crypto.secp256k1.PubKey'
	key: Annotated[PublicKeyData, Field(2)]

@dataclass
class ModeInfo(BaseMessage):

	@dataclass
	class Single(BaseMessage):
		mode: Annotated[SignMode, Field(1)]

	@dataclass
	class Multi(BaseMessage):
		bitarray:  Annotated[bytes, Field(1)]
		modeInfos: Annotated[List[bytes], Field(2)]

	single: Annotated[Optional[Single], Field(1)] = None
	multi:  Annotated[Optional[Multi], Field(2)] = None

@dataclass
class SignerInfo(BaseMessage):
	publicKey: Annotated[PublicKey, Field(1)]
	# mode_info describes the signing mode of the signer and is a nested
	# structure to support nested multisig pubkeys
	modeInfo:  Annotated[ModeInfo, Field(2)]
	# sequence is the sequence of the account, which describes the number of committed
	# transactions signed by a given address. It is used to prevent replay attacks.
	sequence:  Annotated[int, Field(3)]

@dataclass
class Coin(BaseMessage):
	denom:  Annotated[str, Field(1)]
	amount: Annotated[str, Field(2)]

@dataclass
class Fee(BaseMessage):
	amount:   Annotated[Optional[List[Coin]], Field(1)] = None # = field(default_factory=list)
	gasLimit: Annotated[Optional[int], Field(2)] = None
	payer:    Annotated[Optional[str], Field(3)] = None
	granter:  Annotated[Optional[str], Field(4)] = None

@dataclass
class AuthInfo(BaseMessage):
	signerInfos: Annotated[List[SignerInfo], Field(1)]
	fee:         Annotated[Optional[Fee], Field(2)] = None
	tip:         Annotated[Optional[int], Field(3)] = None

# TxRaw is a variant of Tx that pins the signer's exact binary representation of body
# and auth_info. This is used for signing, broadcasting and verification. The binary
# `serialize(tx: TxRaw)` is stored in Tendermint and the hash `sha256(serialize(tx:
# TxRaw))` becomes the "txhash", commonly used as the transaction ID.
@dataclass
class RawTx(BaseMessage): # TxRaw (cosmjs)
	bodyBytes:     Annotated[bytes, Field(1)]
	authInfoBytes: Annotated[bytes, Field(2)]
	signatures:    Annotated[List[bytes], Field(3)]

# SignDoc is the type used for generating sign bytes for SIGN_MODE_DIRECT
@dataclass
class SignDoc(BaseMessage):
	bodyBytes:     Annotated[bytes, Field(1)]
	authInfoBytes: Annotated[bytes, Field(2)]
	chainId:       Annotated[str, Field(3)]
	accountNumber: Annotated[int, Field(4)]

@dataclass
class TxMsg(BaseMessage):
	msgs_cls = None
	id:        Annotated[str, Field(1)]
	bodyBytes: Annotated[bytes, Field(2)]

	def __new__(cls, *, id, bodyBytes):
		msg_cls = getattr(cls.msgs_cls, id.removeprefix('/types.'))
		me = BaseMessage.__new__(msg_cls)
		me.id = id
		me.body = getattr(msg_cls, 'Body').loads(bodyBytes)
		return me

@dataclass
class Tx(BaseMessage):
	body:       Annotated[bytes, Field(1)]
	authInfo:   Annotated[bytes, Field(2)]
	signatures: Annotated[List[bytes], Field(3)]

	@property
	def raw(self):
		return RawTx(
			bodyBytes = bytes(self.body),
			authInfoBytes = bytes(self.authInfo),
			signatures = self.signatures)

	@property
	def txid(self):
		return sha256(bytes(self.raw)).hexdigest()

	# raises exception on failure:
	def verify_sig(self, proto, account_number, backend='secp256k1'):
		sign_doc = SignDoc(
			bodyBytes = bytes(self.body),
			authInfoBytes = bytes(self.authInfo),
			chainId = proto.chain_id,
			accountNumber = account_number)
		sig = self.signatures[0]
		pubkey = self.authInfo.signerInfos[0].publicKey.key.data
		msghash = sha256(bytes(sign_doc)).digest()

		if backend == 'secp256k1':
			from ...secp256k1.secp256k1 import verify_sig
			if not verify_sig(sig, msghash, pubkey):
				raise ValueError('signature verification failed')
		elif backend == 'ecdsa':
			# ecdsa.keys.VerifyingKey.verify_digest():
			#   raises BadSignatureError if the signature is invalid or malformed
			import ecdsa
			ec_pubkey = ecdsa.VerifyingKey.from_string(pubkey, curve=ecdsa.curves.SECP256k1)
			ec_pubkey.verify_digest(sig, msghash)
		else:
			raise ValueError(f'verify_sig(): {backend}: unrecognized backend')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xmr.keygen: Monero public key generation backends for the MMGen suite
"""

from ...key import PubKey
from ...keygen import keygen_base

class backend:

	class base(keygen_base):

		def __init__(self, cfg):
			super().__init__(cfg)
			from ...proto.xmr.params import mainnet
			self.proto_cls = mainnet
			from ...util2 import get_keccak
			self.keccak_256 = get_keccak(cfg)

		def to_viewkey(self, privkey):
			return self.proto_cls.preprocess_key(
				self.proto_cls,
				self.keccak_256(privkey).digest(),
				None)

	class nacl(base):

		production_safe = True

		def __init__(self, cfg):
			super().__init__(cfg)
			from nacl.bindings import crypto_scalarmult_ed25519_base_noclamp
			self.scalarmultbase = crypto_scalarmult_ed25519_base_noclamp

		def to_pubkey(self, privkey):
			return PubKey(
				self.scalarmultbase(privkey) +
				self.scalarmultbase(self.to_viewkey(privkey)),
				compressed = privkey.compressed
			)

	class ed25519(base):

		production_safe = False

		def __init__(self, cfg):
			super().__init__(cfg)
			from ...contrib.ed25519 import edwards, encodepoint, B, scalarmult
			self.edwards     = edwards
			self.encodepoint = encodepoint
			self.B           = B
			self.scalarmult  = scalarmult

		def scalarmultbase(self, privnum):
			"""
			Source and license for scalarmultbase function:
			  https://github.com/bigreddmachine/MoneroPy/blob/master/moneropy/crypto/ed25519.py
			Copyright (c) 2014-2016, The Monero Project
			All rights reserved.
			"""
			if privnum == 0:
				return [0, 1]
			Q = self.scalarmult(self.B, privnum//2)
			Q = self.edwards(Q, Q)
			if privnum & 1:
				Q = self.edwards(Q, self.B)
			return Q

		@staticmethod
		def rev_bytes2int(in_bytes):
			return int.from_bytes(in_bytes[::-1], 'big')

		def to_pubkey(self, privkey):
			return PubKey(
				self.encodepoint(self.scalarmultbase(self.rev_bytes2int(privkey))) +
				self.encodepoint(self.scalarmultbase(self.rev_bytes2int(self.to_viewkey(privkey)))),
				compressed = privkey.compressed
			)

	class ed25519ll_djbec(ed25519):

		def __init__(self, cfg):
			super().__init__(cfg)
			from ...contrib.ed25519ll_djbec import scalarmult
			self.scalarmult = scalarmult



import os as overlay_fake_os
from .unspent_orig import *

if overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA'):

	class overlay_fake_TwUnspentOutputs(TwUnspentOutputs):

		async def set_dates(self, us):
			for o in us:
				o.date = 1831006505 - int(9.7 * 60 * (o.confs - 1))

	TwUnspentOutputs.set_dates = overlay_fake_TwUnspentOutputs.set_dates



#!/usr/bin/env python3
#
# Source: https://github.com/monero-project/monero/blob/master/src/mnemonics/english.h

words = tuple("""
abbey
abducts
ability
ablaze
abnormal
abort
abrasive
absorb
abyss
academy
aces
aching
acidic
acoustic
acquire
across
actress
acumen
adapt
addicted
adept
adhesive
adjust
adopt
adrenalin
adult
adventure
aerial
afar
affair
afield
afloat
afoot
afraid
after
against
agenda
aggravate
agile
aglow
agnostic
agony
agreed
ahead
aided
ailments
aimless
airport
aisle
ajar
akin
alarms
album
alchemy
alerts
algebra
alkaline
alley
almost
aloof
alpine
already
also
altitude
alumni
always
amaze
ambush
amended
amidst
ammo
amnesty
among
amply
amused
anchor
android
anecdote
angled
ankle
annoyed
answers
antics
anvil
anxiety
anybody
apart
apex
aphid
aplomb
apology
apply
apricot
aptitude
aquarium
arbitrary
archer
ardent
arena
argue
arises
army
around
arrow
arsenic
artistic
ascend
ashtray
aside
asked
asleep
aspire
assorted
asylum
athlete
atlas
atom
atrium
attire
auburn
auctions
audio
august
aunt
austere
autumn
avatar
avidly
avoid
awakened
awesome
awful
awkward
awning
awoken
axes
axis
axle
aztec
azure
baby
bacon
badge
baffles
bagpipe
bailed
bakery
balding
bamboo
banjo
baptism
basin
batch
bawled
bays
because
beer
befit
begun
behind
being
below
bemused
benches
berries
bested
betting
bevel
beware
beyond
bias
bicycle
bids
bifocals
biggest
bikini
bimonthly
binocular
biology
biplane
birth
biscuit
bite
biweekly
blender
blip
bluntly
boat
bobsled
bodies
bogeys
boil
boldly
bomb
border
boss
both
bounced
bovine
bowling
boxes
boyfriend
broken
brunt
bubble
buckets
budget
buffet
bugs
building
bulb
bumper
bunch
business
butter
buying
buzzer
bygones
byline
bypass
cabin
cactus
cadets
cafe
cage
cajun
cake
calamity
camp
candy
casket
catch
cause
cavernous
cease
cedar
ceiling
cell
cement
cent
certain
chlorine
chrome
cider
cigar
cinema
circle
cistern
citadel
civilian
claim
click
clue
coal
cobra
cocoa
code
coexist
coffee
cogs
cohesive
coils
colony
comb
cool
copy
corrode
costume
cottage
cousin
cowl
criminal
cube
cucumber
cuddled
cuffs
cuisine
cunning
cupcake
custom
cycling
cylinder
cynical
dabbing
dads
daft
dagger
daily
damp
dangerous
dapper
darted
dash
dating
dauntless
dawn
daytime
dazed
debut
decay
dedicated
deepest
deftly
degrees
dehydrate
deity
dejected
delayed
demonstrate
dented
deodorant
depth
desk
devoid
dewdrop
dexterity
dialect
dice
diet
different
digit
dilute
dime
dinner
diode
diplomat
directed
distance
ditch
divers
dizzy
doctor
dodge
does
dogs
doing
dolphin
domestic
donuts
doorway
dormant
dosage
dotted
double
dove
down
dozen
dreams
drinks
drowning
drunk
drying
dual
dubbed
duckling
dude
duets
duke
dullness
dummy
dunes
duplex
duration
dusted
duties
dwarf
dwelt
dwindling
dying
dynamite
dyslexic
each
eagle
earth
easy
eating
eavesdrop
eccentric
echo
eclipse
economics
ecstatic
eden
edgy
edited
educated
eels
efficient
eggs
egotistic
eight
either
eject
elapse
elbow
eldest
eleven
elite
elope
else
eluded
emails
ember
emerge
emit
emotion
empty
emulate
energy
enforce
enhanced
enigma
enjoy
enlist
enmity
enough
enraged
ensign
entrance
envy
epoxy
equip
erase
erected
erosion
error
eskimos
espionage
essential
estate
etched
eternal
ethics
etiquette
evaluate
evenings
evicted
evolved
examine
excess
exhale
exit
exotic
exquisite
extra
exult
fabrics
factual
fading
fainted
faked
fall
family
fancy
farming
fatal
faulty
fawns
faxed
fazed
feast
february
federal
feel
feline
females
fences
ferry
festival
fetches
fever
fewest
fiat
fibula
fictional
fidget
fierce
fifteen
fight
films
firm
fishing
fitting
five
fixate
fizzle
fleet
flippant
flying
foamy
focus
foes
foggy
foiled
folding
fonts
foolish
fossil
fountain
fowls
foxes
foyer
framed
friendly
frown
fruit
frying
fudge
fuel
fugitive
fully
fuming
fungal
furnished
fuselage
future
fuzzy
gables
gadget
gags
gained
galaxy
gambit
gang
gasp
gather
gauze
gave
gawk
gaze
gearbox
gecko
geek
gels
gemstone
general
geometry
germs
gesture
getting
geyser
ghetto
ghost
giant
giddy
gifts
gigantic
gills
gimmick
ginger
girth
giving
glass
gleeful
glide
gnaw
gnome
goat
goblet
godfather
goes
goggles
going
goldfish
gone
goodbye
gopher
gorilla
gossip
gotten
gourmet
governing
gown
greater
grunt
guarded
guest
guide
gulp
gumball
guru
gusts
gutter
guys
gymnast
gypsy
gyrate
habitat
hacksaw
haggled
hairy
hamburger
happens
hashing
hatchet
haunted
having
hawk
haystack
hazard
hectare
hedgehog
heels
hefty
height
hemlock
hence
heron
hesitate
hexagon
hickory
hiding
highway
hijack
hiker
hills
himself
hinder
hippo
hire
history
hitched
hive
hoax
hobby
hockey
hoisting
hold
honked
hookup
hope
hornet
hospital
hotel
hounded
hover
howls
hubcaps
huddle
huge
hull
humid
hunter
hurried
husband
huts
hybrid
hydrogen
hyper
iceberg
icing
icon
identity
idiom
idled
idols
igloo
ignore
iguana
illness
imagine
imbalance
imitate
impel
inactive
inbound
incur
industrial
inexact
inflamed
ingested
initiate
injury
inkling
inline
inmate
innocent
inorganic
input
inquest
inroads
insult
intended
inundate
invoke
inwardly
ionic
irate
iris
irony
irritate
island
isolated
issued
italics
itches
items
itinerary
itself
ivory
jabbed
jackets
jaded
jagged
jailed
jamming
january
jargon
jaunt
javelin
jaws
jazz
jeans
jeers
jellyfish
jeopardy
jerseys
jester
jetting
jewels
jigsaw
jingle
jittery
jive
jobs
jockey
jogger
joining
joking
jolted
jostle
journal
joyous
jubilee
judge
juggled
juicy
jukebox
july
jump
junk
jury
justice
juvenile
kangaroo
karate
keep
kennel
kept
kernels
kettle
keyboard
kickoff
kidneys
king
kiosk
kisses
kitchens
kiwi
knapsack
knee
knife
knowledge
knuckle
koala
laboratory
ladder
lagoon
lair
lakes
lamb
language
laptop
large
last
later
launching
lava
lawsuit
layout
lazy
lectures
ledge
leech
left
legion
leisure
lemon
lending
leopard
lesson
lettuce
lexicon
liar
library
licks
lids
lied
lifestyle
light
likewise
lilac
limits
linen
lion
lipstick
liquid
listen
lively
loaded
lobster
locker
lodge
lofty
logic
loincloth
long
looking
lopped
lordship
losing
lottery
loudly
love
lower
loyal
lucky
luggage
lukewarm
lullaby
lumber
lunar
lurk
lush
luxury
lymph
lynx
lyrics
macro
madness
magically
mailed
major
makeup
malady
mammal
maps
masterful
match
maul
maverick
maximum
mayor
maze
meant
mechanic
medicate
meeting
megabyte
melting
memoir
menu
merger
mesh
metro
mews
mice
midst
mighty
mime
mirror
misery
mittens
mixture
moat
mobile
mocked
mohawk
moisture
molten
moment
money
moon
mops
morsel
mostly
motherly
mouth
movement
mowing
much
muddy
muffin
mugged
mullet
mumble
mundane
muppet
mural
musical
muzzle
myriad
mystery
myth
nabbing
nagged
nail
names
nanny
napkin
narrate
nasty
natural
nautical
navy
nearby
necklace
needed
negative
neither
neon
nephew
nerves
nestle
network
neutral
never
newt
nexus
nibs
niche
niece
nifty
nightly
nimbly
nineteen
nirvana
nitrogen
nobody
nocturnal
nodes
noises
nomad
noodles
northern
nostril
noted
nouns
novelty
nowhere
nozzle
nuance
nucleus
nudged
nugget
nuisance
null
number
nuns
nurse
nutshell
nylon
oaks
oars
oasis
oatmeal
obedient
object
obliged
obnoxious
observant
obtains
obvious
occur
ocean
october
odds
odometer
offend
often
oilfield
ointment
okay
older
olive
olympics
omega
omission
omnibus
onboard
oncoming
oneself
ongoing
onion
online
onslaught
onto
onward
oozed
opacity
opened
opposite
optical
opus
orange
orbit
orchid
orders
organs
origin
ornament
orphans
oscar
ostrich
otherwise
otter
ouch
ought
ounce
ourselves
oust
outbreak
oval
oven
owed
owls
owner
oxidant
oxygen
oyster
ozone
pact
paddles
pager
pairing
palace
pamphlet
pancakes
paper
paradise
pastry
patio
pause
pavements
pawnshop
payment
peaches
pebbles
peculiar
pedantic
peeled
pegs
pelican
pencil
people
pepper
perfect
pests
petals
phase
pheasants
phone
phrases
physics
piano
picked
pierce
pigment
piloted
pimple
pinched
pioneer
pipeline
pirate
pistons
pitched
pivot
pixels
pizza
playful
pledge
pliers
plotting
plus
plywood
poaching
pockets
podcast
poetry
point
poker
polar
ponies
pool
popular
portents
possible
potato
pouch
poverty
powder
pram
present
pride
problems
pruned
prying
psychic
public
puck
puddle
puffin
pulp
pumpkins
punch
puppy
purged
push
putty
puzzled
pylons
pyramid
python
queen
quick
quote
rabbits
racetrack
radar
rafts
rage
railway
raking
rally
ramped
randomly
rapid
rarest
rash
rated
ravine
rays
razor
react
rebel
recipe
reduce
reef
refer
regular
reheat
reinvest
rejoices
rekindle
relic
remedy
renting
reorder
repent
request
reruns
rest
return
reunion
revamp
rewind
rhino
rhythm
ribbon
richly
ridges
rift
rigid
rims
ringing
riots
ripped
rising
ritual
river
roared
robot
rockets
rodent
rogue
roles
romance
roomy
roped
roster
rotate
rounded
rover
rowboat
royal
ruby
rudely
ruffled
rugged
ruined
ruling
rumble
runway
rural
rustled
ruthless
sabotage
sack
sadness
safety
saga
sailor
sake
salads
sample
sanity
sapling
sarcasm
sash
satin
saucepan
saved
sawmill
saxophone
sayings
scamper
scenic
school
science
scoop
scrub
scuba
seasons
second
sedan
seeded
segments
seismic
selfish
semifinal
sensible
september
sequence
serving
session
setup
seventh
sewage
shackles
shelter
shipped
shocking
shrugged
shuffled
shyness
siblings
sickness
sidekick
sieve
sifting
sighting
silk
simplest
sincerely
sipped
siren
situated
sixteen
sizes
skater
skew
skirting
skulls
skydive
slackens
sleepless
slid
slower
slug
smash
smelting
smidgen
smog
smuggled
snake
sneeze
sniff
snout
snug
soapy
sober
soccer
soda
software
soggy
soil
solved
somewhere
sonic
soothe
soprano
sorry
southern
sovereign
sowed
soya
space
speedy
sphere
spiders
splendid
spout
sprig
spud
spying
square
stacking
stellar
stick
stockpile
strained
stunning
stylishly
subtly
succeed
suddenly
suede
suffice
sugar
suitcase
sulking
summon
sunken
superior
surfer
sushi
suture
swagger
swept
swiftly
sword
swung
syllabus
symptoms
syndrome
syringe
system
taboo
tacit
tadpoles
tagged
tail
taken
talent
tamper
tanks
tapestry
tarnished
tasked
tattoo
taunts
tavern
tawny
taxi
teardrop
technical
tedious
teeming
tell
template
tender
tepid
tequila
terminal
testing
tether
textbook
thaw
theatrics
thirsty
thorn
threaten
thumbs
thwart
ticket
tidy
tiers
tiger
tilt
timber
tinted
tipsy
tirade
tissue
titans
toaster
tobacco
today
toenail
toffee
together
toilet
token
tolerant
tomorrow
tonic
toolbox
topic
torch
tossed
total
touchy
towel
toxic
toyed
trash
trendy
tribal
trolling
truth
trying
tsunami
tubes
tucks
tudor
tuesday
tufts
tugs
tuition
tulips
tumbling
tunnel
turnip
tusks
tutor
tuxedo
twang
tweezers
twice
twofold
tycoon
typist
tyrant
ugly
ulcers
ultimate
umbrella
umpire
unafraid
unbending
uncle
under
uneven
unfit
ungainly
unhappy
union
unjustly
unknown
unlikely
unmask
unnoticed
unopened
unplugs
unquoted
unrest
unsafe
until
unusual
unveil
unwind
unzip
upbeat
upcoming
update
upgrade
uphill
upkeep
upload
upon
upper
upright
upstairs
uptight
upwards
urban
urchins
urgent
usage
useful
usher
using
usual
utensils
utility
utmost
utopia
uttered
vacation
vague
vain
value
vampire
vane
vapidly
vary
vastness
vats
vaults
vector
veered
vegan
vehicle
vein
velvet
venomous
verification
vessel
veteran
vexed
vials
vibrate
victim
video
viewpoint
vigilant
viking
village
vinegar
violin
vipers
virtual
visited
vitals
vivid
vixen
vocal
vogue
voice
volcano
vortex
voted
voucher
vowels
voyage
vulture
wade
waffle
wagtail
waist
waking
wallets
wanted
warped
washing
water
waveform
waxing
wayside
weavers
website
wedge
weekday
weird
welders
went
wept
were
western
wetsuit
whale
when
whipped
whole
wickets
width
wield
wife
wiggle
wildly
winter
wipeout
wiring
wise
withdrawn
wives
wizard
wobbly
woes
woken
wolf
womanly
wonders
woozy
worry
wounded
woven
wrap
wrist
wrong
yacht
yahoo
yanks
yard
yawning
yearbook
yellow
yesterday
yeti
yields
yodel
yoga
younger
yoyo
zapped
zeal
zebra
zero
zesty
zigzags
zinger
zippers
zodiac
zombie
zones
zoom
""".split())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.backends.httplib: httplib RPC backend for the MMGen Project
"""

import json, base64

from ...util import die

from ..util import dmsg_rpc, dmsg_rpc_backend, json_encoder

from .base import base

class httplib(base):
	"""
	Ignores *_PROXY environment vars
	"""
	def __del__(self):
		self.session.close()

	def __init__(self, caller):
		super().__init__(caller)
		import http.client
		self.session = http.client.HTTPConnection(caller.host, caller.port, caller.timeout)
		if caller.auth_type == 'basic':
			auth_str = f'{caller.auth.user}:{caller.auth.passwd}'
			auth_str_b64 = 'Basic ' + base64.b64encode(auth_str.encode()).decode()
			self.http_hdrs.update({'Host': self.host, 'Authorization': auth_str_b64})
			dmsg_rpc(f'    RPC AUTHORIZATION data ==> raw: [{auth_str}]\n{"":>31}enc: [{auth_str_b64}]\n')

	async def run(self, payload, timeout, host_path):
		dmsg_rpc_backend(self.host_url, host_path, payload)

		if timeout:
			import http.client
			s = http.client.HTTPConnection(self.host, self.port, timeout)
		else:
			s = self.session

		try:
			s.request(
				method  = 'POST',
				url     = host_path,
				body    = json.dumps(payload, cls=json_encoder),
				headers = self.http_hdrs)
			r = s.getresponse() # => http.client.HTTPResponse instance
		except Exception as e:
			die('RPCFailure', str(e))

		if timeout:
			ret = (r.read(), r.status)
			s.close()
			return ret
		else:
			return (r.read(), r.status)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.include.test_init: Initialization module for test scripts
"""

import sys, os
from pathlib import PurePath
os.environ['MMGEN_TEST_SUITE'] = '1'
repo_root = str(PurePath(*PurePath(__file__).parts[:-3]))
os.chdir(repo_root)
sys.path[0] = repo_root

from test.overlay import overlay_setup
overlay_root = overlay_setup(repo_root)
os.environ['PYTHONPATH'] = overlay_root

if 'TMUX' in os.environ:
	del os.environ['TMUX']

if os.getenv('MMGEN_DEVTOOLS'):
	from mmgen.devinit import init_dev
	init_dev()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.words: MMGen mnemonic wallet class
"""

from .mnemonic import wallet

class wallet(wallet):

	desc = 'MMGen native mnemonic data'
	mn_type = 'MMGen native'
	wl_id = 'mmgen'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
objmethods: Mixin classes for MMGen data objects
"""

import unicodedata
from . import color as color_mod

if 'MMGenObjectDevTools' in __builtins__: # added to builtins by devinit.init_dev()
	MMGenObject = __builtins__['MMGenObjectDevTools']
else:
	class MMGenObject:
		'placeholder - overridden when testing'
		def immutable_attr_init_check(self):
			pass

def truncate_str(s, width): # width = screen width
	wide_count = 0
	for n, ch in enumerate(s, 1):
		wide_count += unicodedata.east_asian_width(ch) in ('F', 'W')
		if n + wide_count > width:
			return s[:n-1] + ('', ' ')[
				unicodedata.east_asian_width(ch) in ('F', 'W')
				and n + wide_count == width + 1]
	raise ValueError('string requires no truncating')

class Hilite:

	color = 'red'
	width = 0
	trunc_ok = True

	# class method equivalent of fmt()
	@classmethod
	def fmtc(cls, s, width, /, *, color=False):
		if len(s) > width:
			assert cls.trunc_ok, "If 'trunc_ok' is false, 'width' must be >= width of string"
			return cls.colorize(s[:width].ljust(width), color=color)
		else:
			return cls.colorize(s.ljust(width), color=color)

	@classmethod
	def hlc(cls, s, *, color=True):
		return getattr(color_mod, cls.color)(s) if color else s

	@classmethod
	def colorize(cls, s, *, color=True):
		return getattr(color_mod, cls.color)(s) if color else s

	@classmethod
	def colorize2(cls, s, *, color=True, color_override=''):
		return getattr(color_mod, color_override or cls.color)(s) if color else s

class HiliteStr(str, Hilite):

	# supports single-width characters only
	def fmt(self, width, /, *, color=False):
		if len(self) > width:
			assert self.trunc_ok, "If 'trunc_ok' is false, 'width' must be >= width of string"
			return self.colorize(self[:width].ljust(width), color=color)
		else:
			return self.colorize(self.ljust(width), color=color)

	# an alternative to fmt(), with double-width char support and other features
	def fmt2(
			self,
			width,                  # screen width - must be at least 2 (one wide char)
			/,
			*,
			color          = False,
			encl           = '',    # if set, must be exactly 2 single-width chars
			nullrepl       = '',
			append_chars   = '',    # single-width chars only
			append_color   = False,
			color_override = ''):

		if self == '':
			return getattr(color_mod, self.color)(nullrepl.ljust(width)) if color else nullrepl.ljust(width)

		s_wide_count = len(['' for ch in self if unicodedata.east_asian_width(ch) in ('F', 'W')])

		a, b = encl or ('', '')
		add_len = len(append_chars) + len(encl)

		if len(self) + s_wide_count + add_len > width:
			assert self.trunc_ok, "If 'trunc_ok' is false, 'width' must be >= screen width of string"
			s = a + (truncate_str(self, width-add_len) if s_wide_count else self[:width-add_len]) + b
		else:
			s = a + self + b

		if append_chars:
			return (
				self.colorize(s, color=color)
				+ self.colorize2(
					append_chars.ljust(width-len(s)-s_wide_count),
					color_override = append_color))
		else:
			return self.colorize2(s.ljust(width-s_wide_count), color=color, color_override=color_override)

	def hl(self, *, color=True):
		return getattr(color_mod, self.color)(self) if color else self

	# an alternative to hl(), with enclosure and color override
	# can be called as an unbound method with class as first argument
	def hl2(self, s=None, *, color=True, encl='', color_override=''):
		if encl:
			return self.colorize2(encl[0]+(s or self)+encl[1], color=color, color_override=color_override)
		else:
			return self.colorize2((s or self), color=color, color_override=color_override)

class InitErrors:

	@classmethod
	def init_fail(cls, e, m, *, e2=None, m2=None, objname=None, preformat=False):

		def get_errmsg():
			ret = m if preformat else (
				'{!r}: value cannot be converted to {} {}({!s})'.format(
					m,
					(objname or cls.__name__),
					(f'({e2!s}) ' if e2 else ''),
					e))
			return f'{m2!r}\n{ret}' if m2 else ret

		if hasattr(cls, 'passthru_excs') and type(e).__name__ in cls.passthru_excs:
			raise e

		from .util import die
		die(getattr(cls, 'exc', 'ObjectInitError'), get_errmsg())

	@classmethod
	def method_not_implemented(cls):
		import traceback
		raise NotImplementedError(
			'method {}() not implemented for class {!r}'.format(
				traceback.extract_stack()[-2].name, cls.__name__))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
mmgen-cli: Communicate with a coin daemon via its JSON-RPC interface
"""

import asyncio, json

from .util2 import cliargs_convert
from .cfg import gc, Config
from .rpc import rpc_init
from .rpc.util import json_encoder

opts_data = {
	'text': {
		'desc':    'Communicate with a coin daemon via its JSON-RPC interface',
		'usage':   '[opts] <command> <command args>',
		'options': """
-h, --help             Print this help message
--, --longhelp         Print help message for long (global) options
-a, --ascii-output     Ensure that output is ASCII encoded
-w, --wallet=NAME      Use tracking wallet with name NAME
""",
	'notes': """

The utility accepts all {pn} global configuration options and sources the user
config file, allowing users to preconfigure hosts, ports, passwords, datadirs,
tracking wallets and so forth, thus saving a great deal of typing at the
command line. This behavior may be overridden with the --skip-cfg-file option.

Arguments are given in JSON format, with lowercase ‘true’, ‘false’ and ‘null’
for booleans and None, and double-quoted strings in dicts and lists.


                                   EXAMPLES

  $ mmgen-cli --wallet=wallet2 listreceivedbyaddress 0 true

  $ mmgen-cli --coin=ltc --rpc-host=orion getblockcount

  $ mmgen-cli --regtest=1 --wallet=bob getbalance

  $ mmgen-cli --coin=eth eth_getBalance 0x00000000219ab540356cBB839Cbe05303d7705Fa latest

  $ mmgen-cli createrawtransaction \\
    '[{{"txid":"832f5aa9af55dc453314e26869c8f96db1f2a9acac9f23ae18d396903971e0c6","vout":0}}]' \\
    '[{{"1111111111111111111114oLvT2":0.001}}]'
"""
	},
	'code': {
		'notes': lambda cfg, s, help_notes: s.format(
			pn = gc.proj_name)
	}
}

cfg = Config(opts_data=opts_data)
cmd, *args = cfg._args

async def main():

	c = await rpc_init(cfg)
	ret = await c.call(cmd, *cliargs_convert(args), wallet=cfg.wallet)
	print(
		(ascii(ret) if cfg.ascii_output else ret) if isinstance(ret, str) else
		json.dumps(ret, cls=json_encoder, indent=4, ensure_ascii=cfg.ascii_output))

asyncio.run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.util: various utilities for Ethereum base protocol
"""

from ...util2 import get_keccak

v_base = 27

def decrypt_geth_keystore(cfg, wallet_fn, passwd, *, check_addr=True):
	"""
	Decrypt the encrypted private key in a Geth keystore wallet, returning the decrypted key
	"""
	import json

	with open(wallet_fn) as fp:
		wallet_data = json.loads(fp.read())

	from ...altcoin.util import decrypt_keystore
	key = decrypt_keystore(
		wallet_data,
		passwd,
		mac_algo = get_keccak())

	# Optionally check that Ethereum private key produces correct address
	if check_addr:
		from ...tool.coin import tool_cmd
		from ...protocol import init_proto
		t = tool_cmd(cfg=cfg, proto=init_proto(cfg, 'eth'))
		addr = t.wif2addr(key.hex())
		addr_chk = wallet_data['address']
		assert addr == addr_chk, f'incorrect address: ({addr} != {addr_chk})'

	return key

def hash_message(cfg, message, msghash_type):
	return get_keccak(cfg)(
		{
			'raw': message,
			'eth_sign': '\x19Ethereum Signed Message:\n{}{}'.format(len(message), message),
		}[msghash_type].encode()
	).digest()

def ec_sign_message_with_privkey(cfg, message, key, msghash_type):
	"""
	Sign an arbitrary string with an Ethereum private key, returning the signature

	Conforms to the standard defined by the Geth `eth_sign` JSON-RPC call
	"""
	from ..secp256k1.secp256k1 import sign_msghash
	sig, recid = sign_msghash(hash_message(cfg, message, msghash_type), key)
	return sig.hex() + '{:02x}'.format(v_base + recid)

def ec_recover_pubkey(cfg, message, sig, msghash_type):
	"""
	Given a message and signature, recover the public key associated with the private key
	used to make the signature

	Conforms to the standard defined by the Geth `eth_sign` JSON-RPC call
	"""
	from ..secp256k1.secp256k1 import pubkey_recover
	sig_bytes = bytes.fromhex(sig)
	return pubkey_recover(
		hash_message(cfg, message, msghash_type),
		sig_bytes[:64],
		sig_bytes[64] - v_base,
		False).hex()

def compute_contract_addr(cfg, deployer_addr, nonce):
	from . import rlp
	encoded = rlp.encode([bytes.fromhex(deployer_addr), nonce])
	return get_keccak(cfg)(encoded).hexdigest()[-40:]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.submit: Monero wallet ops for the MMGen Suite
"""

import time
from pathlib import Path

from ...util import msg, msg_r, gmsg, die
from ...ui import keypress_confirm
from ...proto.xmr.daemon import MoneroWalletDaemon
from ...proto.xmr.rpc import MoneroWalletRPCClient

from ..file.tx import MoneroMMGenTX
from ..rpc import MoneroWalletRPC

from . import OpBase
from .wallet import OpWallet

class OpSubmit(OpWallet):
	action = 'submitting transaction with'
	opts = ('tx_relay_daemon',)

	def post_mount_action(self):
		self.tx # trigger an exit if no suitable transaction present

	@property
	def tx(self):
		if not hasattr(self, '_tx'):
			self._tx = self.get_tx()
		return self._tx

	def get_tx(self):
		if self.uargs.infile:
			fn = Path(self.uargs.infile)
		else:
			from ...autosign import Signable
			fn = Signable.xmr_transaction(self.asi).get_unsubmitted()
		return MoneroMMGenTX.ColdSigned(cfg=self.cfg, fn=fn)

	def get_relay_rpc(self):

		relay_opt = self.parse_tx_relay_opt()

		wd = MoneroWalletDaemon(
			cfg         = self.cfg,
			proto       = self.proto,
			wallet_dir  = self.cfg.wallet_dir or '.',
			test_suite  = self.cfg.test_suite,
			monerod_addr = relay_opt[1],
		)

		u = wd.usr_daemon_args = []
		if self.cfg.test_suite:
			u.append('--daemon-ssl-allow-any-cert')
		if relay_opt[2]:
			u.append(f'--proxy={relay_opt[2]}')

		return MoneroWalletRPCClient(
			cfg             = self.cfg,
			daemon          = wd,
			test_connection = False,
		)

	async def main(self):
		tx = self.tx
		h = MoneroWalletRPC(self, self.kal.entry(tx.src_wallet_idx))
		self.head_msg(tx.src_wallet_idx, h.fn)
		h.open_wallet()

		if self.cfg.tx_relay_daemon:
			await self.c.stop_daemon()
			self.c = self.get_relay_rpc()
			self.c.start_daemon()
			h = MoneroWalletRPC(self, self.kal.entry(tx.src_wallet_idx))
			h.open_wallet('TX-relay-configured watch-only', refresh=False)

		msg('\n' + tx.get_info(indent='    '))

		if self.cfg.tx_relay_daemon:
			self.display_tx_relay_info(indent='    ')

		keypress_confirm(self.cfg, f'{self.name.capitalize()} transaction?', do_exit=True)

		if self.cfg.tx_relay_daemon:
			msg_r('Relaying transaction to remote daemon, please be patient...')
			t_start = time.time()
		res = self.c.call(
			'submit_transfer',
			tx_data_hex = tx.data.signed_txset)
		assert res['tx_hash_list'][0] == tx.data.txid, 'TxID mismatch in ‘submit_transfer’ result!'
		if self.cfg.tx_relay_daemon:
			from ...util2 import format_elapsed_hr
			msg(f'success\nRelay time: {format_elapsed_hr(t_start, rel_now=False, show_secs=True)}')

		new_tx = MoneroMMGenTX.NewSubmitted(
			cfg          = self.cfg,
			_in_tx       = tx,
		)
		gmsg('\nOK')
		new_tx.write(
			ask_write     = not self.cfg.autosign,
			ask_overwrite = not self.cfg.autosign)
		return new_tx

class OpResubmit(OpSubmit):
	action = 'resubmitting transaction with'

	def check_uopts(self):
		if not self.cfg.autosign:
			die(1, '--autosign is required for this operation')

	def get_tx(self):
		from ...autosign import Signable
		fns = Signable.xmr_transaction(self.asi).get_submitted()
		return sorted(
			(MoneroMMGenTX.Submitted(self.cfg, Path(fn)) for fn in fns),
				key = lambda x: getattr(x.data, 'submit_time', None) or x.data.create_time
		)[-1]

class OpAbort(OpBase):
	opts = ('watch_only', 'autosign')

	def __init__(self, cfg, uarg_tuple):
		super().__init__(cfg, uarg_tuple)
		self.mount_removable_device()
		from ...autosign import Signable
		Signable.xmr_transaction(self.asi).shred_abortable() # prompts user, then raises exception or exits



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.wallet: xmrwallet wallet op for the MMGen Suite
"""

import asyncio, re, atexit
from pathlib import Path

from ...color import orange
from ...util import msg, gmsg, ymsg, die, suf
from ...addrlist import KeyAddrList, ViewKeyAddrList, AddrIdxList
from ...proto.xmr.rpc import MoneroRPCClient, MoneroWalletRPCClient
from ...proto.xmr.daemon import MoneroWalletDaemon

from . import OpBase

class OpWallet(OpBase):

	opts = (
		'use_internal_keccak_module',
		'hash_preset',
		'daemon',
		'no_start_wallet_daemon',
		'no_stop_wallet_daemon',
		'autosign',
		'watch_only',
	)
	wallet_offline = False
	wallet_exists = True
	start_daemon = True
	skip_wallet_check = False # for debugging

	def __init__(self, cfg, uarg_tuple):

		def wallet_exists(fn):
			try:
				fn.stat()
			except:
				return False
			else:
				return True

		def check_wallets():
			for d in self.addr_data:
				fn = self.get_wallet_fn(d)
				exists = wallet_exists(fn)
				if exists and not self.wallet_exists:
					die(1, f'Wallet ‘{fn}’ already exists!')
				elif not exists and self.wallet_exists:
					die(1, f'Wallet ‘{fn}’ not found!')

		super().__init__(cfg, uarg_tuple)

		if self.cfg.offline or (self.name == 'create' and self.cfg.restore_height is None):
			self.wallet_offline = True

		self.wd = MoneroWalletDaemon(
			cfg         = self.cfg,
			proto       = self.proto,
			wallet_dir  = self.cfg.wallet_dir or '.',
			test_suite  = self.cfg.test_suite,
			monerod_addr = self.cfg.daemon or None,
			trust_monerod = self.trust_monerod,
			test_monerod = not self.wallet_offline,
		)

		if self.wallet_offline:
			self.wd.usr_daemon_args = ['--offline']

		self.c = MoneroWalletRPCClient(
			cfg             = self.cfg,
			daemon          = self.wd,
			test_connection = False,
		)

		if self.cfg.offline:
			from ...wallet import Wallet
			self.seed_src = Wallet(
				cfg           = cfg,
				fn            = self.uargs.infile,
				ignore_in_fmt = True)

			gmsg('\nCreating ephemeral key-address list for offline wallets')
			self.kal = KeyAddrList(
				cfg       = cfg,
				proto     = self.proto,
				seed      = self.seed_src.seed,
				addr_idxs = self.uargs.wallets,
				skip_chksum_msg = True)
		else:
			self.mount_removable_device()
			# with watch_only, make a second attempt to open the file as KeyAddrList:
			for first_try in (True, False):
				try:
					self.kal = (ViewKeyAddrList if (self.cfg.watch_only and first_try) else KeyAddrList)(
						cfg    = cfg,
						proto  = self.proto,
						infile = str(self.autosign_viewkey_addr_file) if self.cfg.autosign else self.uargs.infile,
						key_address_validity_check = True,
						skip_chksum_msg = True)
					break
				except:
					if first_try:
						msg(f'Attempting to open ‘{self.uargs.infile}’ as key-address list')
						continue
					raise

		self.create_addr_data()

		if not self.skip_wallet_check:
			check_wallets()

		if self.start_daemon and not self.cfg.no_start_wallet_daemon:
			asyncio.run(self.restart_wallet_daemon())

	@classmethod
	def get_idx_from_fn(cls, fn):
		return int(re.match(r'[0-9a-fA-F]{8}-(\d+)-Monero(WatchOnly)?Wallet.*', fn.name)[1])

	def pre_init_action(self):
		if self.cfg.skip_empty_accounts:
			msg(orange('Skipping display of empty accounts where applicable'))
		if self.cfg.skip_empty_addresses:
			msg(orange('Skipping display of empty used addresses where applicable'))

	def get_coin_daemon_rpc(self):

		host, port = self.cfg.daemon.split(':') if self.cfg.daemon else ('localhost', self.wd.monerod_port)

		from ...daemon import CoinDaemon
		return MoneroRPCClient(
			cfg    = self.cfg,
			proto  = self.proto,
			daemon = CoinDaemon(self.cfg, network_id='xmr'),
			host   = host,
			port   = int(port),
			user   = None,
			passwd = None)

	@property
	def autosign_viewkey_addr_file(self):
		from ...addrfile import ViewKeyAddrFile
		flist = [f for f in self.asi.xmr_dir.iterdir() if f.name.endswith(ViewKeyAddrFile.ext)]
		if len(flist) != 1:
			die(2,
				'{a} viewkey-address files found in autosign mountpoint directory ‘{b}’!\n'.format(
					a = 'Multiple' if flist else 'No',
					b = self.asi.xmr_dir
				)
				+ 'Have you run ‘mmgen-autosign setup’ on your offline machine with the --xmrwallets option?'
			)
		else:
			return flist[0]

	def create_addr_data(self):
		if self.uargs.wallets:
			idxs = AddrIdxList(fmt_str=self.uargs.wallets)
			self.addr_data = [d for d in self.kal.data if d.idx in idxs]
			if len(self.addr_data) != len(idxs):
				die(1, f'List {self.uargs.wallets!r} contains addresses not present in supplied key-address file')
		else:
			self.addr_data = self.kal.data

	async def restart_wallet_daemon(self):
		atexit.register(lambda: asyncio.run(self.stop_wallet_daemon()))
		await self.c.restart_daemon()

	async def stop_wallet_daemon(self):
		if not self.cfg.no_stop_wallet_daemon:
			try:
				await self.c.stop_daemon()
			except KeyboardInterrupt:
				ymsg('\nForce killing wallet daemon')
				self.c.daemon.force_kill = True
				self.c.daemon.stop()

	def get_wallet_fn(self, data, *, watch_only=None):
		if watch_only is None:
			watch_only = self.cfg.watch_only
		return Path(
			(self.cfg.wallet_dir or '.'),
			'{a}-{b}-Monero{c}Wallet{d}'.format(
				a = self.kal.al_id.sid,
				b = data.idx,
				c = 'WatchOnly' if watch_only else '',
				d = f'.{self.cfg.network}' if self.cfg.network != 'mainnet' else '')
		)

	@property
	def add_wallet_desc(self):
		return 'offline signing ' if self.cfg.offline else 'watch-only ' if self.cfg.watch_only else ''

	async def main(self):
		gmsg('\n{a}ing {b} {c}wallet{d}'.format(
			a = self.stem.capitalize(),
			b = len(self.addr_data),
			c = self.add_wallet_desc,
			d = suf(self.addr_data)))
		processed = 0
		for n, d in enumerate(self.addr_data): # [d.sec,d.addr,d.wallet_passwd,d.viewkey]
			fn = self.get_wallet_fn(d)
			gmsg('\n{a}ing wallet {b}/{c} ({d})'.format(
				a = self.stem.capitalize(),
				b = n + 1,
				c = len(self.addr_data),
				d = fn.name,
			))
			processed += await self.process_wallet(d, fn, last=n==len(self.addr_data)-1)
		gmsg(f'\n{processed} wallet{suf(processed)} {self.stem}ed\n')
		return processed

	def head_msg(self, wallet_idx, fn):
		gmsg('\n{a} {b}wallet #{c} ({d})'.format(
			a = self.action.capitalize(),
			b = self.add_wallet_desc,
			c = wallet_idx,
			d = fn.name
		))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tw.ctl: THORChain tracking wallet control class
"""

from ....util import ymsg
from ....tw.store import TwCtlWithStore

class THORChainTwCtl(TwCtlWithStore):

	use_cached_balances = True

	async def rpc_get_balance(self, addr, block='latest'):
		assert self.rpc.is_remote, 'tw.store.rpc_get_balance(): RPC is not remote!'
		try:
			return self.rpc.get_balance(addr, block=block)
		except Exception as e:
			ymsg(f'{type(e).__name__}: {e}')
			ymsg(f'Unable to get balance for address ‘{addr}’')
			if self.get_cached_balance(addr, self.cur_balances, self.data_root):
				from ....ui import keypress_confirm
				if keypress_confirm(self.cfg, 'Zero the balance for this address?'):
					return self.proto.coin_amt('0')
			else:
				import asyncio
				await asyncio.sleep(3)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.sync: Monero wallet ops for the MMGen Suite
"""

import time

from ...util import msg, msg_r, ymsg, die

from ..rpc import MoneroWalletRPC

from .wallet import OpWallet

class OpSync(OpWallet):
	opts = ('rescan_blockchain', 'skip_empty_accounts', 'skip_empty_addresses')

	def check_uopts(self):
		if self.cfg.rescan_blockchain and self.cfg.watch_only:
			die(1, f'Operation ‘{self.name}’ does not support --rescan-blockchain with watch-only wallets')

	def __init__(self, cfg, uarg_tuple):

		super().__init__(cfg, uarg_tuple)

		if not self.wallet_offline:
			self.dc = self.get_coin_daemon_rpc()

		self.wallets_data = {}

	async def process_wallet(self, d, fn, last):

		chain_height = self.dc.call_raw('get_height')['height']
		msg(f'  Chain height: {chain_height}')

		t_start = time.time()

		msg_r('  Opening wallet...')
		self.c.call(
			'open_wallet',
			filename = fn.name,
			password = d.wallet_passwd)
		msg('done')

		msg_r('  Getting wallet height (be patient, this could take a long time)...')
		wallet_height = self.c.call('get_height')['height']
		msg_r('\r' + ' '*68 + '\r')
		msg(f'  Wallet height: {wallet_height}        ')

		behind = chain_height - wallet_height
		if behind > 1000:
			msg_r(f'  Wallet is {behind} blocks behind chain tip.  Please be patient.  Syncing...')

		ret = self.c.call('refresh')

		if behind > 1000:
			msg('done')

		if ret['received_money']:
			msg('  Wallet has received funds')

		for i in range(2):
			wallet_height = self.c.call('get_height')['height']
			if wallet_height >= chain_height:
				break
			ymsg(f'  Wallet failed to sync (wallet height [{wallet_height}] < chain height [{chain_height}])')
			if i or not self.cfg.rescan_blockchain:
				break
			msg_r('  Rescanning blockchain, please be patient...')
			self.c.call('rescan_blockchain')
			self.c.call('refresh')
			msg('done')

		t_elapsed = int(time.time() - t_start)

		wd = MoneroWalletRPC(self, d).get_wallet_data(print=False, skip_empty_ok=True)

		from . import hl_amt
		msg('  Balance: {} Unlocked balance: {}'.format(
			hl_amt(wd.accts_data['total_balance']),
			hl_amt(wd.accts_data['total_unlocked_balance']),
		))

		self.wallets_data[fn.name] = wd

		msg(f'  Wallet height: {wallet_height}')
		msg(f'  Sync time: {t_elapsed//60:02}:{t_elapsed%60:02}')

		if not last:
			self.c.call('close_wallet')

		return wallet_height >= chain_height

	def gen_body(self, wallets_data):
		for wnum, (_, wallet_data) in enumerate(wallets_data.items()):
			yield from MoneroWalletRPC(self, self.addr_data[wnum]).gen_accts_info(
				wallet_data.accts_data,
				wallet_data.addrs_data,
				indent = '',
				skip_empty_ok = True)
			yield ''

	def post_main_success(self):

		def gen_info(data):
			yield from self.gen_body(data)

			col1_w = max(map(len, data)) + 1
			fs = '{:%s} {} {}' % col1_w
			tbals = [0, 0]
			yield fs.format('Wallet', 'Balance           ', 'Unlocked Balance')

			from . import fmt_amt
			for k in data:
				b  = data[k].accts_data['total_balance']
				ub = data[k].accts_data['total_unlocked_balance']
				yield fs.format(k + ':', fmt_amt(b), fmt_amt(ub))
				tbals[0] += b
				tbals[1] += ub

			yield fs.format('-'*col1_w, '-'*18, '-'*18)
			yield fs.format('TOTAL:', fmt_amt(tbals[0]), fmt_amt(tbals[1]))

		self.cfg._util.stdout_or_pager('\n'.join(gen_info(self.wallets_data)) + '\n')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.httpd.etherscan: Etherscan WSGI http server
"""

from mmgen.util2 import get_keccak

from . import HTTPD

class EtherscanServer(HTTPD):
	name = 'etherscan server'
	port = 28800
	content_type = 'text/html'

	def make_response_body(self, method, environ):
		if method == 'GET':
			target = 'form'
		elif method == 'POST':
			target = 'result'
			length = int(environ.get('CONTENT_LENGTH', '0'))
			qs = environ['wsgi.input'].read(length).decode()
			tx = [s for s in qs.split('&') if 'RawTx=' in s][0].split('=')[1]
			keccak_256 = get_keccak()
			txid = '0x' + keccak_256(bytes.fromhex(tx[2:])).hexdigest()

		with open(f'test/ref/ethereum/etherscan-{target}.html') as fh:
			text = fh.read()

		return (text if method == 'GET' else text.format(txid=txid)).encode()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.rpc: Bitcoin base protocol tracking wallet RPC classes
"""

from ....addr import CoinAddr
from ....util import die, msg
from ....tw.shared import get_tw_label
from ....tw.rpc import TwRPC
from ....tw.ctl import label_addr_pair

class BitcoinTwRPC(TwRPC):

	async def get_label_addr_pairs(self):
		"""
		Get all the accounts in the tracking wallet and their associated addresses.
		Returns list of (label, address) tuples.
		"""
		def check_dup_mmid(acct_labels):
			mmid_prev, err = None, False
			for mmid in sorted(label.mmid for label in acct_labels if label):
				if mmid == mmid_prev:
					err = True
					msg(f'Duplicate MMGen ID ({mmid}) discovered in tracking wallet!\n')
				mmid_prev = mmid
			if err:
				die(4, 'Tracking wallet is corrupted!')

		async def get_acct_list():
			if 'label_api' in self.rpc.caps:
				return await self.rpc.call('listlabels')
			else:
				return (await self.rpc.call('listaccounts', 0, True)).keys()

		async def get_acct_addrs(acct_list):
			if 'label_api' in self.rpc.caps:
				return [list(a.keys())
					for a in await self.rpc.batch_call('getaddressesbylabel', [(k,) for k in acct_list])]
			else:
				return await self.rpc.batch_call('getaddressesbyaccount', [(a,) for a in acct_list])

		acct_labels = [get_tw_label(self.proto, a) for a in await get_acct_list()]

		if not acct_labels:
			return []

		check_dup_mmid(acct_labels)

		acct_addrs = await get_acct_addrs(acct_labels)

		for n, a in enumerate(acct_addrs):
			if len(a) != 1:
				raise ValueError(f'{a}: label {acct_labels[n]!r} has != 1 associated address!')

		return [label_addr_pair(label, CoinAddr(self.proto, addrs[0]))
			for label, addrs in zip(acct_labels, acct_addrs)]

	async def get_unspent_by_mmid(self, *, minconf=1, mmid_filter=[]):
		"""
		get unspent outputs in tracking wallet, compute balances per address
		and return a dict with elements {'twmmid': {'addr', 'lbl', 'amt'}}
		"""
		data = {}
		lbl_id = ('account', 'label')['label_api' in self.rpc.caps]
		amt0 = self.proto.coin_amt('0')

		for d in await self.rpc.call('listunspent', 0):

			if not lbl_id in d:
				continue  # skip coinbase outputs with missing account

			if d['confirmations'] < minconf:
				continue

			label = get_tw_label(self.proto, d[lbl_id])

			if label:
				lm = label.mmid
				if mmid_filter and (lm not in mmid_filter):
					continue
				if lm in data:
					if data[lm]['addr'] != d['address']:
						die(2, 'duplicate {} address ({}) for this MMGen address! ({})'.format(
							self.proto.coin,
							d['address'],
							data[lm]['addr']))
				else:
					lm.confs = d['confirmations']
					lm.txid = d['txid']
					lm.vout = d['vout']
					lm.date = None
					data[lm] = {
						'amt': amt0,
						'lbl': label,
						'addr': CoinAddr(self.proto, d['address'])}

				data[lm]['amt'] += self.proto.coin_amt(d['amount'])

		return data



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.params: Ethereum protocol
"""

from ...protocol import CoinProtocol, _nw, decoded_addr
from ...addr import CoinAddr
from ...util import is_hex_str_lc, Msg

class mainnet(CoinProtocol.DummyWIF, CoinProtocol.Secp256k1):

	network_names = _nw('mainnet', 'testnet', 'devnet')
	addr_len      = 20
	mmtypes       = ('E',)
	preferred_mmtypes  = ('E',)
	dfl_mmtype    = 'E'
	mod_clsname   = 'Ethereum'
	pubkey_type   = 'std' # required by DummyWIF

	coin_amt      = 'ETHAmt'
	max_tx_fee    = 0.005
	chain_names   = ['ethereum', 'foundation']
	sign_mode     = 'standalone'
	caps          = ('token',)
	mmcaps        = ('rpc', 'rpc_init', 'tw', 'msg')
	base_proto    = 'Ethereum'
	base_proto_coin = 'ETH'
	base_coin     = 'ETH'
	avg_bdi       = 15
	decimal_prec  = 36
	address_reuse_ok = True
	is_vm = True
	is_evm = True

	# https://www.chainid.dev
	chain_ids = {
		1:    'ethereum',         # ethereum mainnet
		2:    'morden',           # morden testnet (deprecated)
		3:    'ropsten',          # ropsten testnet
		4:    'rinkeby',          # rinkeby testnet
		5:    'goerli',           # goerli testnet
		42:   'kovan',            # kovan testnet
		61:   'classic',          # ethereum classic mainnet
		62:   'morden',           # ethereum classic testnet
		17:   'developmentchain', # parity dev chain
		1337: 'developmentchain', # geth dev chain
		711:  'ethereum',         # geth mainnet (empty chain)
		17000: 'holesky',         # proof-of-stake testnet
	}

	coin_cfg_opts = (
		'daemon_id',
		'ignore_daemon_version',
		'rpc_host',
		'rpc_port',
		'max_tx_fee',
	)

	proto_cfg_opts = (
		'chain_names',
	)

	@property
	def dcoin(self):
		return self.tokensym or self.coin

	def decode_addr(self, addr):
		if is_hex_str_lc(addr) and len(addr) == self.addr_len * 2:
			return decoded_addr(bytes.fromhex(addr), None, 'p2pkh')
		if self.cfg.debug:
			Msg(f'Invalid address: {addr}')
		return False

	def checksummed_addr(self, addr):
		h = self.keccak_256(addr.encode()).digest().hex()
		return ''.join(addr[i].upper() if int(h[i], 16) > 7 else addr[i] for i in range(len(addr)))

	def pubhash2addr(self, pubhash, addr_type):
		assert len(pubhash) == 20, f'{len(pubhash)}: invalid length for {self.name} pubkey hash'
		assert addr_type == 'p2pkh', (
				f'{addr_type}: bad addr type - {self.name} protocol supports P2PKH address format only')
		return CoinAddr(self, pubhash.hex())

class testnet(mainnet):
	chain_names = ['kovan', 'goerli', 'rinkeby']

class regtest(testnet):
	chain_names = ['developmentchain']



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.tx.new: Ethereum new transaction class
"""

from ....tx import new as TxBase
from ....obj import Int, ETHNonce
from ....util import msg, ymsg, is_int

from ...vm.tx.new import New as VmNew

from ..contract import Token

from .base import Base, TokenBase

class New(VmNew, Base, TxBase.New):
	fee_fail_fs = 'Network fee estimation failed'
	usr_fee_prompt = 'Enter transaction fee or gas price: '
	byte_cost = 68 # https://ethereum.stackexchange.com/questions/39401/
	               # how-do-you-calculate-gas-limit-for-transaction-with-data-in-ethereum

	def __init__(self, *args, **kwargs):

		super().__init__(*args, **kwargs)

		if self.cfg.contract_data:
			m = "'--contract-data' option may not be used with token transaction"
			assert 'Token' not in self.name, m
			with open(self.cfg.contract_data) as fp:
				self.usr_contract_data = bytes.fromhex(fp.read().strip())
			self.disable_fee_check = True

	async def get_gas_estimateGas(self, *, to_addr):
		return self.dfl_gas

	async def set_gas(self, *, to_addr=None, force=False):
		if force or to_addr or not hasattr(self, 'gas'):
			if is_int(self.cfg.gas):
				self.gas = int(self.cfg.gas)
			elif self.cfg.gas == 'fallback':
				self.gas = self.dfl_gas
			elif self.is_bump and not self.rpc.daemon.id == 'reth':
				self.gas = self.txobj['startGas']
			else:
				assert self.cfg.gas in ('auto', None), f'{self.cfg.gas}: invalid value for cfg.gas'
				self.gas = await self.get_gas_estimateGas(to_addr=to_addr)

	async def get_nonce(self):
		return ETHNonce(int(
			await self.rpc.call('eth_getTransactionCount', '0x'+self.inputs[0].addr, 'pending'), 16))

	async def make_txobj(self): # called by create_serialized()
		self.txobj = {
			'from': self.inputs[0].addr,
			'to':   self.outputs[0].addr if self.outputs else None,
			'amt':  self.outputs[0].amt if self.outputs else self.proto.coin_amt('0'),
			'gasPrice': self.fee_abs2gasprice(self.usr_fee),
			'startGas': self.gas,
			'nonce': await self.get_nonce(),
			'chainId': self.rpc.chainID,
			'data':  self.usr_contract_data.hex()}

	def set_gas_with_data(self, data):
		if not self.is_token:
			self.gas = self.dfl_gas + self.byte_cost * len(data)

	# one-shot method
	def adj_gas_with_extra_data_len(self, extra_data_len):
		if not (self.is_token or hasattr(self, '_gas_adjusted')):
			self.gas += self.byte_cost * extra_data_len
			self._gas_adjusted = True

	@property
	def network_estimated_fee_label(self):
		return 'Network-estimated'

	def network_fee_to_unit_disp(self, net_fee):
		return '{} Gwei'.format(self.pretty_fmt_fee(
			self.proto.coin_amt(net_fee.fee, from_unit='wei').to_unit('Gwei')))

	# get rel_fee (gas price) from network, return in native wei
	async def get_rel_fee_from_network(self):
		return self._net_fee(
			Int(await self.rpc.call('eth_gasPrice'), base=16),
			'eth_gasPrice')

	# given rel fee and units, return absolute fee using self.total_gas
	def fee_rel2abs(self, tx_size, amt_in_units, unit):
		return self.proto.coin_amt(int(amt_in_units * self.total_gas), from_unit=unit)

	# given fee estimate (gas price) in wei, return absolute fee, adjusting by self.cfg.fee_adjust
	def fee_est2abs(self, net_fee):
		ret = self.fee_gasPrice2abs(net_fee.fee) * self.cfg.fee_adjust
		if self.cfg.verbose:
			msg(f'Estimated fee: {net_fee.fee} ETH')
		return ret

	def convert_and_check_fee(self, fee, desc):
		abs_fee = self.feespec2abs(fee, None)
		if abs_fee is False:
			return False
		elif not self.disable_fee_check and (abs_fee > self.proto.max_tx_fee):
			msg('{} {c}: {} fee too large (maximum fee: {} {c})'.format(
				abs_fee.hl(),
				desc,
				self.proto.max_tx_fee.hl(),
				c = self.proto.coin))
			return False
		else:
			return abs_fee

class TokenNew(TokenBase, New):
	desc = 'transaction'
	fee_is_approximate = True

	async def set_gas(self, *, to_addr=None, force=False):
		await super().set_gas(to_addr=to_addr, force=force)
		if self.is_swap and (force or not hasattr(self, 'router_gas')):
			self.router_gas = (
				int(self.cfg.router_gas) if self.cfg.router_gas else
				self.txobj['router_gas'] if self.txobj else
				self.dfl_router_gas)

	@property
	def total_gas(self):
		return self.gas + (self.router_gas if self.is_swap else 0)

	async def get_gas_estimateGas(self, *, to_addr=None):
		t = Token(
			self.cfg,
			self.proto,
			self.twctl.token,
			decimals = self.twctl.decimals,
			rpc = self.rpc)

		data = t.create_transfer_data(
			to_addr = to_addr or self.outputs[0].addr,
			amt = self.outputs[0].amt or await self.twctl.get_balance(self.inputs[0].addr),
			op = self.token_op)

		try:
			res = await t.do_call(
				f'{self.token_op}(address,uint256)',
				method = 'eth_estimateGas',
				from_addr = self.inputs[0].addr,
				data = data)
		except Exception as e:
			ymsg(
				'Unable to estimate gas limit via node. '
				'Please retry with --gas set to an integer value, or ‘fallback’ for a sane default')
			raise e

		return int(res, 16)

	async def make_txobj(self): # called by create_serialized()
		await super().make_txobj()
		t = Token(self.cfg, self.proto, self.twctl.token, decimals=self.twctl.decimals)
		o = self.txobj
		o['token_addr'] = t.addr
		o['decimals'] = t.decimals
		o['token_to'] = o['to']
		if self.is_swap:
			o['expiry'] = self.quote_data.data['expiry']
			o['router_gas'] = self.router_gas

	def update_change_output(self, funds_left):
		if self.outputs[0].is_chg:
			self.update_output_amt(0, self.inputs[0].amt)

	# token transaction, so check both eth and token balances
	# TODO: add test with insufficient funds
	async def precheck_sufficient_funds(self, inputs_sum, sel_unspent, outputs_sum):
		eth_bal = await self.twctl.get_eth_balance(sel_unspent[0].addr)
		if eth_bal == 0: # we don't know the fee yet
			msg('This account has no ether to pay for the transaction fee!')
			return False
		return await super().precheck_sufficient_funds(inputs_sum, sel_unspent, outputs_sum)

	async def get_funds_available(self, fee, outputs_sum):
		bal = await self.twctl.get_eth_balance(self.inputs[0].addr)
		return self._funds_available(bal >= fee, bal - fee if bal >= fee else fee - bal)

	def final_inputs_ok_msg(self, funds_left):
		token_bal = (
			self.proto.coin_amt('0') if self.outputs[0].is_chg
			else self.inputs[0].amt - self.outputs[0].amt
		)
		return "Transaction leaves ≈{} {} and {} {} in the sender's account".format(
			funds_left.hl(),
			self.proto.coin,
			token_bal.hl(),
			self.proto.dcoin
		)
